from __future__ import annotations

__all__ = ["is_parameter", "is_uninitialized_parameter"]

from typing import Any

from torch.nn import Parameter, UninitializedParameter


def is_parameter(value: Any) -> bool:
    r"""Indicates if the input value is a ``torch.nn.Parameter``.

    Args:
    ----
        value: Specifies the value to check.

    Returns:
    -------
        bool: ``True`` if the input value is a ``torch.nn.Parameter``,
            otherwise ``False``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.nn import Parameter, UninitializedParameter
        >>> from gravitorch.utils.parameter import is_parameter
        >>> is_parameter(Parameter(torch.ones(2, 3)))
        True
        >>> is_parameter(UninitializedParameter())
        True
        >>> is_parameter(torch.ones(2, 3))
        False
    """
    return isinstance(value, Parameter)


def is_uninitialized_parameter(value: Any) -> bool:
    r"""Indicates if the input value is a
    ``torch.nn.UninitializedParameter``.

    Args:
    ----
        value: Specifies the value to check.

    Returns:
    -------
        bool: ``True`` if the input value is a
            ``torch.nn.UninitializedParameter``, otherwise ``False``.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.nn import Parameter, UninitializedParameter
        >>> from gravitorch.utils.parameter import is_uninitialized_parameter
        >>> is_parameter(UninitializedParameter())
        True
        >>> is_parameter(Parameter(torch.ones(2, 3)))
        False
        >>> is_parameter(torch.ones(2, 3))
        False
    """
    return isinstance(value, UninitializedParameter)



__all__ = ["SequentialInitializer"]

import logging
from collections.abc import Sequence
from typing import Union

from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer
from gravitorch.nn.init.factory import setup_initializer
from gravitorch.utils.format import str_indent, str_torch_sequence

logger = logging.getLogger(__name__)


class SequentialInitializer(BaseInitializer):
    r"""Implements a module initializer that sequentially calls module
    initializers.

    Args:
    ----
        initializers: Specifies the sequence of module initializers.
            The sequence order defines the order of the call.
    """

    def __init__(
        self,
        initializers: Sequence[Union[BaseInitializer, dict]],
    ) -> None:
        super().__init__()
        self._initializers = tuple(setup_initializer(initializer) for initializer in initializers)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_torch_sequence(self._initializers))}\n)"
        )

    def initialize(self, module: Module) -> None:
        for i, initializer in enumerate(self._initializers):
            logger.info(f"[{i}/{len(self._initializers)}] {initializer}")
            initializer.initialize(module)



from __future__ import annotations

__all__ = [
    "BaseDataSource",
    "LoaderNotFoundError",
    "is_datasource_config",
    "setup_datasource",
    "setup_and_attach_datasource",
]

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataSource(ABC, Generic[T], metaclass=AbstractFactory):
    r"""Defines the base class to implement a datasource.

    A datasource object is responsible to create the data loaders.

    Note: it is an experimental class and the API may change.
    """

    @abstractmethod
    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the current datasource to the provided engine.

        This method can be used to set up events or logs some stats to
        the engine.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            >>> datasource.attach(my_engine)
        """

    @abstractmethod
    def get_asset(self, asset_id: str) -> Any:
        r"""Gets a data asset from this datasource.

        This method is useful to access some data variables/parameters
        that are not available before to load/preprocess the data.

        Args:
        ----
            asset_id (str): Specifies the ID of the asset.

        Returns:
        -------
            The asset.

        Raises:
        ------
            ``AssetNotFoundError`` if you try to access an asset that
                does not exist.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            >>> my_asset = datasource.get_asset("my_asset_id")
        """

    @abstractmethod
    def has_asset(self, asset_id: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
        ----
            asset_id (str): Specifies the ID of the asset.

        Returns:
        -------
            bool: ``True`` if the asset exists, otherwise ``False``.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            >>> datasource.has_asset("my_asset_id")
            False
        """

    @abstractmethod
    def get_dataloader(self, loader_id: str, engine: BaseEngine | None = None) -> Iterable[T]:
        r"""Gets a data loader.

        Args:
        ----
            loader_id (str): Specifies the ID of the data loader to
                get.
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                data loader by using the current epoch value.
                Default: ``None``

        Returns:
        -------
            ``Iterable``: A data loader.

        Raises:
        ------
            ``LoaderNotFoundError`` if the loader does not exist.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            # Get the data loader associated to the ID 'train'
            >>> dataloader = datasource.get_dataloader("train")
            # Get a data loader that can use information from an engine
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> dataloader = datasource.get_dataloader("train", my_engine)
        """

    @abstractmethod
    def has_dataloader(self, loader_id: str) -> bool:
        r"""Indicates if the datasource has a data loader with the given
        ID.

        Args:
        ----
            loader_id (str): Specifies the ID of the data loader.

        Returns:
        -------
            bool: ``True`` if the data loader exists, ``False``
                otherwise.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            # Check if the datasource has a data loader for ID 'train'
            >>> datasource.has_dataloader("train")
            True or False
            # Check if the datasource has a data loader for ID 'eval'
            >>> datasource.has_dataloader("eval")
            True or False
        """

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
        ----
            state_dict (dict): a dict with parameters

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            # Please take a look to the implementation of the state_dict
            # function to know the expected structure
            >>> state_dict = {...}
            >>> datasource.load_state_dict(state_dict)
        """

    def state_dict(self) -> dict:
        r"""Returns a dictionary containing state values.

        Returns:
        -------
            dict: the state values in a dict.

        Example:
        -------
        .. code-block:: pycon

            >>> from gravitorch.datasources import BaseDataSource
            >>> datasource: BaseDataSource = ...  # Instantiate a datasource
            >>> state_dict = datasource.state_dict()
            {...}
        """
        return {}


class LoaderNotFoundError(Exception):
    r"""Raised when a loader is requested but does not exist."""


def is_datasource_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataSource``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataSource`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datasources import is_datasource_config
        >>> is_datasource_config(
        ...     {"_target_": "gravitorch.datasources.IterDataPipeCreatorDataSource"}
        ... )
        True
    """
    return is_object_config(config, BaseDataSource)


def setup_datasource(datasource: BaseDataSource | dict) -> BaseDataSource:
    r"""Sets up a datasource.

    The datasource is instantiated from its configuration by using
    the ``BaseDataSource`` factory function.

    Args:
    ----
        datasource (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.

    Returns:
    -------
        ``BaseDataSource``: The instantiated datasource.
    """
    if isinstance(datasource, dict):
        logger.info(
            "Initializing a datasource from its configuration... "
            f"{str_target_object(datasource)}"
        )
        datasource = BaseDataSource.factory(**datasource)
    if not isinstance(datasource, BaseDataSource):
        logger.warning(f"datasource is not a `BaseDataSource` (received: {type(datasource)})")
    return datasource


def setup_and_attach_datasource(
    datasource: BaseDataSource | dict, engine: BaseEngine
) -> BaseDataSource:
    r"""Sets up a datasource and attach it to an engine.

    Note that if you call this function ``N`` times with the same data
    source object, the datasource will be attached ``N`` times to the
    engine.

    Args:
    ----
        datasource (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.
        engine (``BaseEngine``): Specifies the engine.

    Returns:
    -------
        ``BaseDataSource``: The instantiated datasource.
    """
    datasource = setup_datasource(datasource)
    logger.info("Adding a datasource object to an engine...")
    datasource.attach(engine)
    return datasource



__all__ = ["IterableBatchLoader"]

from collections.abc import Iterable, Iterator
from types import TracebackType
from typing import Optional, TypeVar

from gravitorch.experimental.batchloader.base import BaseBatchLoader

T = TypeVar("T")


class IterableBatchLoader(BaseBatchLoader[T]):
    r"""Implements a simple batch loader that works on any iterable.

    Args:
        iterable (``Iterable``): Specifies the iterable of batches.
    """

    def __init__(self, iterable: Iterable) -> None:
        self._iterable = iterable

    def __enter__(self) -> "IterableBatchLoader":
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def __iter__(self) -> Iterator[T]:
        return iter(self._iterable)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(iterable={self._iterable})"



from __future__ import annotations

__all__ = [
    "AutoDataLoaderCreator",
    "DataLoaderCreator",
    "DistributedDataLoaderCreator",
]

from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import DataLoader, Dataset, DistributedSampler

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.dataloaders.factory import create_dataloader
from gravitorch.distributed import comm as dist
from gravitorch.utils.format import str_indent, str_pretty_dict
from gravitorch.utils.seed import get_torch_generator

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class AutoDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a PyTorch data loader creator that automatically chooses
    the data loader creator based on the context.

    If the distributed package is activated, it uses the
    ``DistributedDataLoaderCreator``, otherwise it uses
    ``DataLoaderCreator``.


    Note the behavior of this class may change based on the new data
    loader creators.

    Args:
    ----
        **kwargs: See ``DataLoaderCreator`` or
            ``DistributedDataLoaderCreator`` documentation.
    """

    def __init__(self, **kwargs) -> None:
        if dist.is_distributed():
            self._creator = DistributedDataLoaderCreator(**kwargs)
        else:
            self._creator = DataLoaderCreator(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  creator={str_indent(str(self._creator))},\n)"

    def create(self, dataset: Dataset, engine: BaseEngine | None = None) -> DataLoader[T]:
        return self._creator.create(dataset=dataset, engine=engine)


class DataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a simple PyTorch data loader creator.

    Note that this data loader creator uses the default samplers.
    If you need a different sampler, you will need to implement your
    own data loader creator.

    Args:
    ----
        seed (int, optional): Specifies the random seed used to
            reproduce the shuffling of the samples. Default: ``0``
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.
    """

    def __init__(self, seed: int = 0, **kwargs) -> None:
        self._seed = int(seed)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {"seed": self._seed} | self._kwargs
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_pretty_dict(config, sorted_keys=True))}\n)"
        )

    def create(self, dataset: Dataset, engine: BaseEngine | None = None) -> DataLoader[T]:
        epoch = 0 if engine is None else engine.epoch
        return create_dataloader(
            dataset, generator=get_torch_generator(self._seed + epoch), **self._kwargs
        )


class DistributedDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a simple distributed PyTorch data loader creator.

    This data loader creator uses the ``gravitorch.distributed`` package
    to distribute the example per process. Note that this data loader
    creator uses the default samplers. If you need a different sampler,
    you will need to implement your own data loader creator.

    Args:
    ----
        shuffle (bool, optional): Specifies of the examples are
            shuffled or not. You should set to ``True`` to have the
            data reshuffled at every epoch. Default: ``False``
        drop_last (bool, optional): set to ``True`` to drop the last
            incomplete batch, if the dataset size is not divisible by
            the batch size. If ``False`` and the size of dataset is
            not divisible by the batch size, then the last batch will
            be smaller. Default: ``False``
        seed (int, optional): Specifies the random seed used to
            shuffle the samples if ``shuffle=True``. Default: ``0``
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.
    """

    def __init__(
        self, shuffle: bool = True, drop_last: bool = False, seed: int = 0, **kwargs
    ) -> None:
        self._shuffle = bool(shuffle)
        self._drop_last = bool(drop_last)
        self._seed = int(seed)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {
            "shuffle": self._shuffle,
            "drop_last": self._drop_last,
            "seed": self._seed,
        } | self._kwargs
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_pretty_dict(config, sorted_keys=True))}\n)"
        )

    def create(self, dataset: Dataset, engine: BaseEngine | None = None) -> DataLoader[T]:
        sampler = DistributedSampler(
            dataset,
            shuffle=self._shuffle,
            drop_last=self._drop_last,
            seed=self._seed,
            rank=dist.get_rank(),
            num_replicas=dist.get_world_size(),
        )
        epoch = 0
        if engine is not None:
            epoch = engine.epoch
            # In distributed mode, calling the set_epoch() method at the beginning
            # of each epoch before creating the DataLoader iterator is necessary to
            # make shuffling work properly across multiple epochs.
            # Otherwise, the same ordering will always be used.
            sampler.set_epoch(epoch)

        # Sampler option is mutually exclusive with shuffle or drop last.
        return create_dataloader(
            dataset,
            sampler=sampler,
            generator=get_torch_generator(self._seed + epoch),
            **self._kwargs,
        )



__all__ = ["ModelFreezer"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.events import VanillaEventHandler
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.nn import freeze_module

logger = logging.getLogger(__name__)


class ModelFreezer(BaseHandler):
    r"""Implements a handler to freeze the model or one of its
    submodules.

    Args:
    ----
        event (str, optional): Specifies the event when the model
            is frozen. Default: ``'train_started'``
        module_name (str, optional): Specifies the name of the module
            to freeze if only one of the submodules should be frozen.
            Default: ``''``
    """

    def __init__(
        self,
        event: str = EngineEvents.TRAIN_STARTED,
        module_name: str = "",
    ) -> None:
        self._module_name = str(module_name)
        self._event = str(event)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, module_name={self._module_name})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self.freeze,
                handler_kwargs={"engine": engine},
            ),
        )

    def freeze(self, engine: BaseEngine) -> None:
        r"""Freezes the model or one of its submodules.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine with the
                model.
        """
        if self._module_name:
            logger.info(f"Freeze submodule {self._module_name}")
        else:
            logger.info("Freeze model")
        freeze_module(engine.model.get_submodule(self._module_name))



__all__ = [
    "AMPEvaluationLoop",
    "AccelerateEvaluationLoop",
    "BaseBasicEvaluationLoop",
    "BaseEvaluationLoop",
    "NoOpEvaluationLoop",
    "VanillaEvaluationLoop",
    "is_evaluation_loop_config",
    "setup_evaluation_loop",
]

from gravitorch.loops.evaluation.accelerate import AccelerateEvaluationLoop
from gravitorch.loops.evaluation.amp import AMPEvaluationLoop
from gravitorch.loops.evaluation.base import BaseEvaluationLoop
from gravitorch.loops.evaluation.basic import BaseBasicEvaluationLoop
from gravitorch.loops.evaluation.factory import (
    is_evaluation_loop_config,
    setup_evaluation_loop,
)
from gravitorch.loops.evaluation.noop import NoOpEvaluationLoop
from gravitorch.loops.evaluation.vanilla import VanillaEvaluationLoop



r"""This module defines some utility functions to format some
objects."""

from __future__ import annotations

__all__ = [
    "human_byte_size",
    "human_count",
    "human_time",
    "str_indent",
    "str_mapping",
    "str_pretty_dict",
    "str_pretty_json",
    "str_pretty_yaml",
    "str_scalar",
    "str_target_object",
    "str_torch_mapping",
    "str_torch_sequence",
]

import datetime
import json
import math
from collections.abc import Mapping, Sequence
from typing import Any, TypeVar

import yaml
from objectory import OBJECT_TARGET

PARAMETER_NUM_UNITS = (" ", "K", "M", "B", "T")

BYTE_UNITS = {
    "B": 1,
    "KB": 1024,
    "MB": 1024 * 1024,
    "GB": 1024 * 1024 * 1024,
    "TB": 1024 * 1024 * 1024 * 1024,
}

T = TypeVar("T")


def human_byte_size(size: int, unit: str | None = None) -> str:
    r"""Gets a human-readable representation of the byte size.

    Args:
    ----
        size (int): Specifies the size in bytes.
        unit (str, optional): Specifies the unit. If ``None``, the
            best unit is found automatically. The supported units
            are: ``'B'``, ``'KB'``, ``'MB'``, ``'GB'``, ``'TB'``.
            Default: ``None``

    Returns:
    -------
        str: The byte size in a human-readable format.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import human_byte_size
        >>> human_byte_size(2)
        2.00 B
        >>> human_byte_size(2048)
        2.00 KB
        >>> human_byte_size(2097152)
        2.00 MB
        >>> human_byte_size(2048, unit="B")
        2,048.00 B
    """
    if unit is None:  # Find the best unit.
        best_unit = "B"
        for unit, multiplier in BYTE_UNITS.items():
            if (size / multiplier) > 1:
                best_unit = unit
        unit = best_unit

    if unit not in BYTE_UNITS:
        raise ValueError(
            f"Incorrect unit '{unit}'. The available units are {list(BYTE_UNITS.keys())}"
        )

    return f"{size / BYTE_UNITS.get(unit, 1):,.2f} {unit}"


def human_count(number: int | float) -> str:
    r"""Converts an integer number with K, M, B, T for thousands,
    millions, billions and trillions, respectively.

    Args:
    ----
        number (int or float): A positive integer number. If the
            number is a float, it will be converted to an integer.

    Returns:
    -------
        str: A string formatted according to the pattern described
            above.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import human_count
        >>> human_count(123)
        '123  '
        >>> human_count(1234)  # (one thousand)
        '1.2 K'
        >>> human_count(2e6)  # (two million)
        '2.0 M'
        >>> human_count(3e9)  # (three billion)
        '3.0 B'
        >>> human_count(4e14)  # (four hundred trillion)
        '400 T'
        >>> human_count(5e15)  # (more than trillion)
        '5,000 T'
    """
    if number < 0:
        raise ValueError(f"The number should be a positive number (received {number})")
    if number < 1000:
        return str(int(number))
    labels = PARAMETER_NUM_UNITS
    num_digits = int(math.floor(math.log10(number)) + 1 if number > 0 else 1)
    num_groups = min(
        int(math.ceil(num_digits / 3)), len(labels)
    )  # don't abbreviate beyond trillions
    shift = -3 * (num_groups - 1)
    number = number * (10**shift)
    index = num_groups - 1
    if index < 1 or number >= 100:
        return f"{int(number):,d} {labels[index]}"
    return f"{number:,.1f} {labels[index]}"


def human_time(seconds: int | float) -> str:
    r"""Converts a number of seconds in an easier format to read
    hh:mm:ss.

    If the number of seconds is bigger than 1 day, this representation
    also encodes the number of days.

    Args:
    ----
        seconds (integer or float): Specifies the number of seconds.

    Returns:
    -------
        str: The number of seconds in a string format (hh:mm:ss).

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import human_time
        >>> human_time(1.2)
        '0:00:01.200000'
        >>> human_time(61.2)
        '0:01:01.200000'
        >>> human_time(3661.2)
        '1:01:01.200000'
    """
    return str(datetime.timedelta(seconds=seconds))


def str_indent(original: Any, num_spaces: int = 2) -> str:
    r"""Adds indentations if the original string is a multi-lines string.

    Args:
    ----
        original: Specifies the original string. If the inputis not a
            string, it will be converted to a string with the function
            ``str``.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The indented string.

    Example usage:

    .. code-block:: pycon

        >>> string_to_print = "string1\nstring2"
        >>> print(f"{string_to_print}")
        string1
        string2
        >>> print(f"\t{string_to_print}")
            string1
        string2
        # The problem is that 'string1' and 'string2' are not aligned.
        # The indentation is only applied to the first line
        >>> from gravitorch.utils.format import str_indent
        >>> print(f"\t{str_indent(string_to_print, 4)}")
            string1
            string2
    """
    formatted_str = str(original).split("\n")
    if len(formatted_str) == 1:  # single line
        return formatted_str[0]
    first = formatted_str.pop(0)
    formatted_str = "\n".join([(num_spaces * " ") + line for line in formatted_str])
    return first + "\n" + formatted_str


def str_mapping(
    mapping: Mapping,
    sorted_keys: bool = False,
    num_spaces: int = 2,
    one_line: bool = False,
) -> str:
    r"""Computes a string representation of a mapping.

    Args:
    ----
        mapping (``Mapping``): Specifies the mapping.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. This option is used only when
            ``one_line=False``. Default: ``2``.
        one_line (bool, optional): If ``True``, tryes to generate a
            single line string representation. The keys and values
            should not contain multiple lines. If ``False``, a new
            line is created for each key in the input mapping.
            Default: ``False``

    Returns:
    -------
        str: The string representation of the mapping.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_mapping
        >>> str_mapping({"key1": "value1", "key2": "value2"})
        key1=value1
        key2=value2
        >>> str_mapping({"key1": "long\nvalue1", "key2": "value2"})
        key1=long
          value1
        key2=value2
        >>> str_mapping({"key1": "value1", "key2": "value2"}, one_line=True)
        key1=value1, key2=value2
    """
    items = sorted(mapping.items()) if sorted_keys else mapping.items()
    if one_line:
        return ", ".join([f"{key}={value}" for key, value in items])
    return "\n".join([f"{key}={str_indent(value, num_spaces=num_spaces)}" for key, value in items])


def str_scalar(value: int | float) -> str:
    r"""Returns a string representation of a scalar value.

    Args:
    ----
        value (int or float): Specifies the input value.

    Returns:
    -------
        str: The string representation of the input value.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_scalar
        >>> str_scalar(123456.789)
        123,456.789000
        >>> str_scalar(1234567)
        1,234,567
        >>> str_scalar(12345678901)
        1.234568e+10
    """
    if isinstance(value, int):
        if math.fabs(value) >= 1e9:
            return f"{value:.6e}"
        return f"{value:,}"
    if math.fabs(value) < 1e-3 or math.fabs(value) >= 1e6:
        return f"{value:.6e}"
    return f"{value:,.6f}"


def str_target_object(config: dict) -> str:
    r"""Gets a string that indicates the target object in the config.

    Args:
    ----
        config (dict): Specifies a config using the ``object_factory``
            library. This dict is expected to have a key
            ``'_target_'`` to indicate the target object.

    Returns:
    -------
        str: A string with the target object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_target_object
        >>> str_target_object({OBJECT_TARGET: "something.MyClass"})
        [_target_: something.MyClass]
        >>> str_target_object({})
        [_target_: N/A]
    """
    return f"[{OBJECT_TARGET}: {config.get(OBJECT_TARGET, 'N/A')}]"


def str_pretty_json(data: Any, sort_keys: bool = True, indent: int = 2, max_len: int = 80) -> str:
    r"""Converts a data structure to a pretty JSON string.

    Args:
    ----
        data: Specifies the input to convert to a pretty JSON string.
        sort_keys (bool, optional): Specifies if the keys are sorted
            or not. Default: ``True``
        indent (int, optional): Specifies the indent. It is a
            non-negative integer. Default: ``2``
        max_len (int, optional): Specifies the maximum length of the
            string representation. If the string representation is
            longer than this length, it is converted to the json
            representation. Default: ``80``

    Returns:
    -------
        str: The string representation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_pretty_json
        >>> str_pretty_json({"my_key": "my_value"})
        "{'my_key': 'my_value'}"
        >>> str_pretty_json(["value1", "value2"])
        "['value1', 'value2']"
        >>> str_pretty_json(["value1", "value2"], max_len=5)
        '[\n  "value1",\n  "value2"\n]'
    """
    str_data = str(data)
    if len(str_data) < max_len:
        return str_data
    return json.dumps(data, sort_keys=sort_keys, indent=indent, default=str)


def str_pretty_yaml(data: Any, sort_keys: bool = True, indent: int = 2, max_len: int = 80) -> str:
    r"""Converts a data structure to a pretty YAML string.

    Args:
    ----
        data: Specifies the input to convert to a pretty YAML string.
        sort_keys (bool, optional): Specifies if the keys are sorted
            or not. Default: ``True``
        indent (int, optional): Specifies the indent. It is a
            non-negative integer. Default: ``2``
        max_len (int, optional): Specifies the maximum length of the
            string representation. If the string representation is
            longer than this length, it is converted to the json
            representation. Default: ``max_len``

    Returns:
    -------
        str: The string representation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_pretty_yaml
        >>> str_pretty_yaml({"my_key": "my_value"})
        "{'my_key': 'my_value'}"
        >>> str_pretty_yaml(["value1", "value2"])
        "['value1', 'value2']"
        >>> str_pretty_yaml(["value1", "value2"], max_len=5)
        '- value1\n- value2\n'
    """
    str_data = str(data)
    if len(str_data) < max_len:
        return str_data
    return yaml.safe_dump(data, sort_keys=sort_keys, indent=indent)


def str_pretty_dict(data: dict[str, Any], sorted_keys: bool = False, indent: int = 0) -> str:
    r"""Converts a dict to a pretty string representation.

    This function was designed for flat dictionary. If you have a
    nested dictionary, you may consider other functions. Note that
    this function works for nested dict but the output may not be
    nice.

    Args:
    ----
        data (dict): Specifies the input dictionary.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        indent (int, optional): Specifies the indentation. The value
            should be greater or equal to 0. Default: ``0``

    Returns:
    -------
        str: The string representation.

        Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_pretty_dict
        >>> str_pretty_dict({"my_key": "my_value"})
        'my_key : my_value'
        >>> str_pretty_dict({"key1": "value1", "key2": "value2"})
        'key1 : value1\nkey2 : value2'
    """
    if indent < 0:
        raise ValueError(f"The indent has to be greater or equal to 0 (received: {indent})")
    if not data:
        return ""

    max_length = max([len(key) for key in data])
    output = []
    for key in sorted(data.keys()) if sorted_keys else data.keys():
        output.append(f"{' ' * indent + str(key) + ' ' * (max_length - len(key))} : {data[key]}")
    return "\n".join(output)


def str_torch_mapping(mapping: Mapping, sorted_keys: bool = False, num_spaces: int = 2) -> str:
    r"""Computes a torch-like (``torch.nn.Module``) string representation
    of a mapping.

    This function was designed for flat dictionary. If you have a
    nested dictionary, you may consider other functions. Note that
    this function works for nested dict but the output may not be
    nice.

    Args:
    ----
        mapping (``Mapping``): Specifies the mapping.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The string representation of the mapping.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_torch_mapping
        >>> str_torch_mapping({"key1": "abc", "key2": "something\nelse"})
        (key1): abc
        (key2): something
          else
    """
    lines = []
    for key, value in sorted(mapping.items()) if sorted_keys else mapping.items():
        lines.append(f"({key}): {str_indent(value, num_spaces=num_spaces)}")
    return "\n".join(lines)


def str_torch_sequence(sequence: Sequence, num_spaces: int = 2) -> str:
    r"""Computes a torch-like (``torch.nn.Module``) string representation
    of a sequence.

    Args:
    ----
        sequence (``Sequence``): Specifies the sequence.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The string representation of the sequence.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.format import str_torch_sequence
        >>> str_torch_sequence(["abc", "something\nelse"])
        (0): abc
        (1): something
          else
    """
    lines = []
    for i, item in enumerate(sequence):
        lines.append(f"({i}): {str_indent(item, num_spaces=num_spaces)}")
    return "\n".join(lines)



__all__ = [
    "DummyClassificationModel",
    "DummyDataSource",
    "DummyDataset",
    "DummyIterableDataset",
    "accelerate_available",
    "create_dummy_engine",
    "cuda_available",
    "distributed_available",
    "gloo_available",
    "nccl_available",
    "pillow_available",
    "psutil_available",
    "tensorboard_available",
    "torchdata_available",
    "torchvision_available",
    "two_gpus_available",
    "matplotlib_available",
]

from gravitorch.testing._pytest import (
    accelerate_available,
    cuda_available,
    distributed_available,
    gloo_available,
    matplotlib_available,
    nccl_available,
    pillow_available,
    psutil_available,
    tensorboard_available,
    torchdata_available,
    torchvision_available,
    two_gpus_available,
)
from gravitorch.testing.dummy import (
    DummyClassificationModel,
    DummyDataset,
    DummyDataSource,
    DummyIterableDataset,
    create_dummy_engine,
)



r"""This package contains the implementation of some datasets."""

from __future__ import annotations

__all__ = [
    "DummyMultiClassDataset",
    "ImageFolderDataset",
    "ExampleDataset",
    "create_datasets",
    "is_dataset_config",
    "log_box_dataset_class",
    "setup_dataset",
]

from gravitorch.data.datasets.dummy import DummyMultiClassDataset
from gravitorch.data.datasets.example import ExampleDataset
from gravitorch.data.datasets.factory import (
    create_datasets,
    is_dataset_config,
    setup_dataset,
)
from gravitorch.data.datasets.image_folder import ImageFolderDataset
from gravitorch.data.datasets.utils import log_box_dataset_class



r"""This package contains the model base class and some implemented
models."""

from __future__ import annotations

__all__ = [
    "BaseModel",
    "VanillaModel",
    "attach_module_to_engine",
    "is_model_config",
    "setup_and_attach_model",
    "setup_model",
]

from gravitorch.models.base import BaseModel
from gravitorch.models.utils import (
    attach_module_to_engine,
    is_model_config,
    setup_and_attach_model,
    setup_model,
)
from gravitorch.models.vanilla import VanillaModel



r"""This module implements datasets that stores all the examples in
memory."""
from __future__ import annotations

__all__ = ["ExampleDataset", "ExampleDatasetEqualityOperator"]

import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Any, TypeVar

import torch
from coola import (
    BaseEqualityOperator,
    BaseEqualityTester,
    EqualityTester,
    objects_are_equal,
)
from torch.utils.data import Dataset

from gravitorch.data.datasets.utils import log_box_dataset_class
from gravitorch.utils.io import load_json, load_pickle
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)

T = TypeVar("T")


class ExampleDataset(Dataset[T]):
    r"""Implements a dataset that stores all the examples in-memory.

    You can use this dataset only if all the examples can fit
    in-memory.

    Args:
    ----
        examples: Specifies the examples of the dataset.
    """

    def __init__(self, examples: Sequence[T]) -> None:
        log_box_dataset_class(self)
        self._examples = tuple(examples)

    def __len__(self) -> int:
        return len(self._examples)

    def __getitem__(self, item: int) -> T:
        return self._examples[item]

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(num_examples={len(self):,})"

    def equal(self, other: Any) -> bool:
        r"""Indicates if two datasets are equal or not.

        Args:
        ----
            other: Specifies the other dataset to compare.

        Returns:
        -------
            bool: ``True`` if the datasets are equal,
                otherwise ``False``

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.datasets import ExampleDataset
            >>> ExampleDataset([1, 2]).equal(ExampleDataset([1, 2]))
            True
            >>> ExampleDataset([1, 2]).equal(ExampleDataset([2, 1]))
            False
        """
        if not isinstance(other, self.__class__):
            return False
        if self is other:
            return True
        return objects_are_equal(self._examples, other._examples)

    @classmethod
    def from_json_file(cls, path: Path | str) -> ExampleDataset:
        r"""Instantiates a dataset with the examples from a JSON file.

        Args:
        ----
            path (``pathlib.Path`` or str): Specifies the path to the
                JSON file.

        Returns:
        -------
            ``ExampleDataset``: An instantiated dataset.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.datasets import ExampleDataset
            >>> dataset = ExampleDataset.from_json_file("/path/to/file.pt")
        """
        return cls(load_json(sanitize_path(path)))

    @classmethod
    def from_pickle_file(cls, path: Path | str) -> ExampleDataset:
        r"""Instantiates a dataset with the examples from a pickle file.

        Args:
        ----
            path (``pathlib.Path`` or str): Specifies the path to the
                pickle file.

        Returns:
        -------
            ``ExampleDataset``: An instantiated dataset.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.datasets import ExampleDataset
            >>> dataset = ExampleDataset.from_pickle_file("/path/to/file.pkl")
        """
        return cls(load_pickle(sanitize_path(path)))

    @classmethod
    def from_pytorch_file(cls, path: Path | str, **kwargs) -> ExampleDataset:
        r"""Instantiates a dataset with the examples from a PyTorch file.

        Args:
        ----
            path (``pathlib.Path`` or str): Specifies the path to the
                PyTorch file.
            **kwargs: See ``torch.load`` documentation.

        Returns:
        -------
            ``ExampleDataset``: An instantiated dataset.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.datasets import ExampleDataset
            >>> dataset = ExampleDataset.from_pytorch_file("/path/to/file.pt")
        """
        return cls(torch.load(sanitize_path(path), **kwargs))


class ExampleDatasetEqualityOperator(BaseEqualityOperator[ExampleDataset]):
    r"""Implements an equality operator for ``ExampleDataset``
    objects."""

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> ExampleDatasetEqualityOperator:
        return self.__class__()

    def equal(
        self,
        tester: BaseEqualityTester,
        object1: ExampleDataset,
        object2: Any,
        show_difference: bool = False,
    ) -> bool:
        if object1 is object2:
            return True
        if not isinstance(object2, ExampleDataset):
            if show_difference:
                logger.info(f"object2 is not a `ExampleDataset` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`ExampleDataset` objects are different\nobject1=\n{object1}\nobject2=\n{object2}"
            )
        return object_equal


if not EqualityTester.has_operator(ExampleDataset):
    EqualityTester.add_operator(
        ExampleDataset, ExampleDatasetEqualityOperator()
    )  # pragma: no cover



r"""This module defines some utility functions for the data loader
creators."""

from __future__ import annotations

__all__ = ["setup_dataloader_creator"]

import logging

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.pytorch import AutoDataLoaderCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_dataloader_creator(
    creator: BaseDataLoaderCreator | dict | None,
) -> BaseDataLoaderCreator:
    r"""Sets up a data loader creator.

    Args:
    ----
        creator (``BaseDataLoaderCreator`` or dict or None):
            Specifies the data loader creator or its configuration.
            If ``None``, a data loader creator will be created
            automatically.

    Returns:
    -------
        ``BaseDataLoaderCreator``: The data loader creator.
    """
    if creator is None:
        creator = AutoDataLoaderCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing a data loader creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataLoaderCreator.factory(**creator)
    return creator



from __future__ import annotations

__all__ = [
    "BaseDataCreator",
    "DataCreator",
    "HypercubeVertexDataCreator",
    "CacheDataCreator",
    "setup_data_creator",
]

from gravitorch.data.datacreators.base import BaseDataCreator, setup_data_creator
from gravitorch.data.datacreators.caching import CacheDataCreator
from gravitorch.data.datacreators.hypercube import HypercubeVertexDataCreator
from gravitorch.data.datacreators.vanilla import DataCreator



from __future__ import annotations

__all__ = ["DataLoaderDataFlow"]

from collections.abc import Iterator

from torch.utils.data import DataLoader

from gravitorch.experimental.dataflow.base import BaseDataFlow
from gravitorch.utils.imports import is_torchdata_available

if is_torchdata_available():
    from torchdata.dataloader2 import DataLoader2
else:
    DataLoader2 = "DataLoader2"  # pragma: no cover


class DataLoaderDataFlow(BaseDataFlow):
    r"""Implements a simple dataflow for PyTorch data loaders.

    Args:
        dataloader (``DataLoader`` or ``DataLoader2``): Specifies the
            data loader.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.utils.data import DataLoader, TensorDataset
        >>> from gravitorch.experimental.dataflow import IterableDataFlow
        >>> dataloader = DataLoader(TensorDataset(torch.arange(10)), batch_size=4)
        >>> with DataLoaderDataFlow(dataloader) as dataflow:
        ...     for batch in dataflow:
        ...         print(batch)  # do something
        ...
    """

    def __init__(self, dataloader: DataLoader | DataLoader2) -> None:
        if not isinstance(
            dataloader, (DataLoader, DataLoader2) if is_torchdata_available() else DataLoader
        ):
            raise TypeError(
                "Incorrect type. Expecting DataLoader or DataLoader2 but "
                f"received {type(dataloader)}"
            )
        self.dataloader = dataloader

    def __iter__(self) -> Iterator:
        return iter(self.dataloader)

    def __repr__(self) -> str:
        try:
            extra = f"length={len(self.dataloader):,}"
        except TypeError:
            extra = ""
        return f"{self.__class__.__qualname__}({extra})"

    def launch(self) -> None:
        r"""Nothing to do for this dataflow."""

    def shutdown(self) -> None:
        if is_torchdata_available() and isinstance(self.dataloader, DataLoader2):
            self.dataloader.shutdown()



r"""This package contains the implementation of a lot of utilities used
in the other packages."""

__all__ = [
    "get_available_devices",
    "manual_seed",
    "move_to_device",
    "setup_object",
    "to_tuple",
]

from gravitorch.utils.device import get_available_devices, move_to_device
from gravitorch.utils.factory import setup_object
from gravitorch.utils.seed import manual_seed
from gravitorch.utils.seq import to_tuple



from __future__ import annotations

__all__ = ["setup_object"]

import logging
from typing import TypeVar

from objectory import factory

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


def setup_object(obj_or_config: T | dict) -> T:
    r"""Sets up an object from its configuration.

    Args:
    ----
        obj_or_config: Specifies the object or its configuration.

    Returns:
    -------
        The instantiated object.

    Example usage:

    .. code-block:: pycon

       >>> from gravitorch.utils import setup_object
       >>> linear = setup_object(
       ...     {"_target_": "torch.nn.Linear", "in_features": 4, "out_features": 6}
       ... )
       >>> linear
       Linear(in_features=4, out_features=6, bias=True)
       >>> setup_object(linear)  # Do nothing because the module is already instantiated
       Linear(in_features=4, out_features=6, bias=True)
    """
    if isinstance(obj_or_config, dict):
        logger.info(
            "Initializing an object from its configuration... "
            f"{str_target_object(obj_or_config)}"
        )
        obj_or_config = factory(**obj_or_config)
    return obj_or_config



__all__ = ["setup_initializer"]

import logging
from typing import Union

from gravitorch.nn.init.base import BaseInitializer
from gravitorch.nn.init.noop import NoOpInitializer
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_initializer(initializer: Union[BaseInitializer, dict, None]) -> BaseInitializer:
    r"""Sets up the module parameter initializer.

    Args:
    ----
        initializer (``BaseInitializer`` or dict or ``None``):
            Specifies the model parameter initializer or its
            configuration. If ``None``, the ``NoOpInitializer``
            is instantiated.

    Returns:
    -------
        ``BaseInitializer``: The instantiated module parameter
            initializer.
    """
    if initializer is None:
        initializer = NoOpInitializer()
    if isinstance(initializer, dict):
        logger.debug(
            "Initializing a module initializer from its configuration... "
            f"{str_target_object(initializer)}"
        )
        initializer = BaseInitializer.factory(**initializer)
    return initializer



r"""This module defines some functionalities to instantiate dynamically
a ``torch.utils.data.Dataset`` object from its configuration."""

from __future__ import annotations

__all__ = ["create_datasets", "is_dataset_config", "setup_dataset"]

import logging
from collections.abc import Hashable, Mapping
from typing import TypeVar

from objectory import factory
from objectory.utils import is_object_config
from torch.utils.data import Dataset

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


def create_datasets(datasets: Mapping[Hashable, Dataset | dict]) -> dict[Hashable, Dataset]:
    r"""Create datasets indexed by dataset split.

    Args:
    ----
        datasets (``Mapping``): Specifies the datasets or their
            configuration. The key is the dataset split name and the
            value is the dataset or its configuration.

    Returns:
    -------
        dict: The instantiated datasets.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datasets import ExampleDataset, create_datasets
        >>> create_datasets(
        ...     {
        ...         "train": ExampleDataset((1, 2, 3)),
        ...         "val": {
        ...             "_target_": "gravitorch.data.datasets.ExampleDataset",
        ...             "examples": (4, 5),
        ...         },
        ...     }
        ... )
        {'train': ExampleDataset(num_examples=3), 'val': ExampleDataset(num_examples=2)}
    """
    return {split: setup_dataset(dataset) for split, dataset in datasets.items()}


def is_dataset_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``torch.utils.data.Dataset``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``torch.utils.data.Dataset`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datasets import is_dataset_config
        >>> is_dataset_config(
        ...     {"_target_": "gravitorch.data.datasets.ExampleDataset", "examples": [1, 2, 1, 3]}
        ... )
        True
    """
    return is_object_config(config, Dataset)


def setup_dataset(dataset: Dataset | dict | None) -> Dataset | None:
    r"""Sets up a dataset.

    Args:
    ----
        dataset (``Dataset`` or dict or ``None``): Specifies the
            dataset or its configuration (dictionary). If a
            configuration is given, a dataset object is instantiated
            from the configuration.

    Returns:
    -------
        ``torch.utils.data.Dataset`` or ``None``: A dataset object or
            ``None`` if there is no dataset.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datasets import setup_dataset
        >>> mnist = setup_dataset(
        ...     {"_target_": "torchvision.datasets.MNIST", "root": "/my/path/", "download": True},
        ... )
        >>> mnist
        Dataset MNIST
            Number of datapoints: 60000
            Root location: tmp
            Split: Train
        >>> setup_dataset(mnist)  # Do nothing because the dataset is already instantiated
        Dataset MNIST
            Number of datapoints: 60000
            Root location: tmp
            Split: Train
        >>> setup_dataset(None)
    """
    if isinstance(dataset, dict):
        logger.info(
            f"Initializing a dataset from its configuration... {str_target_object(dataset)}"
        )
        dataset = factory(**dataset)
    return dataset



from __future__ import annotations

import torch
from pytest import mark
from torch import distributed as tdist

from gravitorch.distributed.comm import Backend, available_backends
from gravitorch.utils.imports import (
    is_accelerate_available,
    is_matplotlib_available,
    is_pillow_available,
    is_tensorboard_available,
    is_torchdata_available,
    is_torchvision_available,
)

cuda_available = mark.skipif(not torch.cuda.is_available(), reason="Requires a device with CUDA")
two_gpus_available = mark.skipif(torch.cuda.device_count() < 2, reason="Requires at least 2 GPUs")
distributed_available = mark.skipif(not tdist.is_available(), reason="Requires PyTorch distributed")
nccl_available = mark.skipif(Backend.NCCL not in available_backends(), reason="Requires NCCL")
gloo_available = mark.skipif(Backend.GLOO not in available_backends(), reason="Requires GLOO")

accelerate_available = mark.skipif(
    not is_accelerate_available(),
    reason=(
        "`accelerate` is not available. Please install `accelerate` if you want to run this test"
    ),
)
matplotlib_available = mark.skipif(
    not is_matplotlib_available(),
    reason=(
        "`matplotlib` is not available. Please install `matplotlib` if you want to run this test"
    ),
)
pillow_available = mark.skipif(
    not is_pillow_available(),
    reason="`pillow` is not available. Please install `pillow` if you want to run this test",
)
psutil_available = mark.skipif(
    not is_pillow_available(),
    reason="`psutil` is not available. Please install `psutil` if you want to run this test",
)
tensorboard_available = mark.skipif(
    not is_tensorboard_available(),
    reason=(
        "`tensorboard` is not available. Please install `tensorboard` if you want "
        "to run this test"
    ),
)
torchdata_available = mark.skipif(
    not is_torchdata_available(),
    reason=(
        "`torchdata` is not available. Please install `torchdata` if you want to run this test"
    ),
)
torchvision_available = mark.skipif(
    not is_torchvision_available(),
    reason=(
        "`torchvision` is not available. Please install `torchvision` if you want "
        "to run this test"
    ),
)



__all__ = ["NoOpInitializer"]

import logging

from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer

logger = logging.getLogger(__name__)


class NoOpInitializer(BaseInitializer):
    r"""This is the special class that does not update the module
    parameters.

    You should use this class if the parameters of the module are
    initialized somewhere else.
    """

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def initialize(self, module: Module) -> None:
        logger.info("The module parameters are not updated")



r"""The data package contains the engine base class and some implemented
engines."""

__all__ = [
    "AlphaEngine",
    "BaseEngine",
    "EngineEvents",
    "is_engine_config",
    "setup_engine",
]

from gravitorch.engines.alpha import AlphaEngine
from gravitorch.engines.base import BaseEngine, is_engine_config, setup_engine
from gravitorch.engines.events import EngineEvents



r"""This module implements a handler to initialize model's
parameters."""

__all__ = ["ModelInitializer"]

import logging
from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.events import VanillaEventHandler
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.nn.init import BaseInitializer, setup_initializer
from gravitorch.utils.format import str_indent

logger = logging.getLogger(__name__)


class ModelInitializer(BaseHandler):
    r"""Implements a handler to initialize the model's parameters.

    This handler uses a ``BaseInitializer`` object to
    initialize model's parameters.

    Args:
    ----
        initializer (``BaseInitializer`` or dict): Specifies the
            model's parameters initializer or its configuration.
        event (str, optional): Specifies the event when to initialize
            the model's parameters. Default: ``'train_started'``
    """

    def __init__(
        self,
        initializer: Union[BaseInitializer, dict],
        event: str = EngineEvents.TRAIN_STARTED,
    ) -> None:
        self._initializer = setup_initializer(initializer)
        self._event = event

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  initializer={str_indent(self._initializer)},"
            f"  event={self._event},"
            ")"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self._initializer.initialize,
                handler_kwargs={"module": engine.model},
            ),
        )



__all__ = [
    "BaseXavier",
    "XavierNormal",
    "XavierUniform",
    "xavier_normal",
    "xavier_uniform",
]

import logging

from torch import nn
from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer

logger = logging.getLogger(__name__)


class BaseXavier(BaseInitializer):
    r"""Implements a module parameter initializer with the Xavier Normal
    or uniform strategy.

    Args:
    ----
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1.0``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``
    """

    def __init__(
        self,
        gain: float = 1.0,
        learnable_only: bool = True,
        log_info: bool = False,
    ) -> None:
        super().__init__()
        self._gain = float(gain)
        self._learnable_only = bool(learnable_only)
        self._log_info = bool(log_info)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(gain={self._gain}, "
            f"learnable_only={self._learnable_only}, log_info={self._log_info})"
        )


class XavierNormal(BaseXavier):
    r"""Implements a module parameter initializer with the Xavier Normal
    strategy."""

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing module parameters with the Xavier Normal strategy (gain={self._gain}, "
            f"learnable_only={self._learnable_only})..."
        )
        xavier_normal(
            module=module,
            gain=self._gain,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


class XavierUniform(BaseXavier):
    r"""Implements a module parameter initializer with the Xavier
    uniform strategy."""

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing module parameters with the Xavier uniform strategy (gain={self._gain}, "
            f"learnable_only={self._learnable_only})..."
        )
        xavier_uniform(
            module=module,
            gain=self._gain,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


def xavier_normal(
    module: nn.Module,
    gain: float = 1.0,
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Initialize the module parameters with the Xavier Normal strategy.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import xavier_normal
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> xavier_normal(net)
    """
    for name, params in module.named_parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            if log_info:
                logger.info(f"Initializing '{name}' with gain {gain} | shape={params.shape}")
            nn.init.xavier_normal_(params.data, gain=gain)


def xavier_uniform(
    module: nn.Module,
    gain: float = 1.0,
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Initializes the module parameters with the Xavier uniform
    strategy.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import xavier_uniform
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> xavier_uniform(net)
    """
    for name, params in module.named_parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            if log_info:
                logger.info(f"Initializing '{name}' with gain {gain} | shape={params.shape}")
            nn.init.xavier_uniform_(params.data, gain=gain)



__all__ = ["setup_clip_grad"]

import logging
from collections.abc import Callable
from typing import Optional

import torch

logger = logging.getLogger(__name__)


def setup_clip_grad(config: Optional[dict]) -> tuple[Optional[Callable], tuple]:
    r"""Sets up the gradient clipping strategy.

    Args:
    ----
        config (dict): Specifies the configuration of the gradient
            clipping strategy.

    Returns:
    -------
        tuple: clip gradient function, clip gradient arguments.

    Raises:
    ------
        RuntimeError: if it is an invalid clipping gradient name.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.training.utils import setup_clip_grad
        >>> setup_clip_grad({"name": "clip_grad_value", "clip_value": 0.5})
        (<function torch.nn.utils.clip_grad.clip_grad_value_(parameters: Union[torch.Tensor, Iterable[torch.Tensor]], clip_value: float, foreach: Optional[bool] = None) -> None>,
         (0.5,))
        >>> setup_clip_grad({"name": "clip_grad_norm", "max_norm": 0.5, "norm_type": 2.0})
        (<function torch.nn.utils.clip_grad.clip_grad_norm_(parameters: Union[torch.Tensor, Iterable[torch.Tensor]], max_norm: float, norm_type: float = 2.0, error_if_nonfinite: bool = False, foreach: Optional[bool] = None) -> torch.Tensor>,
         (0.5, 2.0))
        >>> setup_clip_grad({})
        (None, ())
    """
    config = config or {}
    if not config:
        return None, tuple()

    name = config["name"]
    if name == "clip_grad_value":
        clip_value = float(config.get("clip_value", 0.25))
        logger.info(f"clip gradient by value {clip_value}")
        return torch.nn.utils.clip_grad_value_, (clip_value,)
    if name == "clip_grad_norm":
        max_norm = float(config.get("max_norm", 1.0))
        norm_type = float(config.get("norm_type", 2.0))
        logger.info(f"clip gradient by maximum norm {max_norm} (norm type: {norm_type})")
        return torch.nn.utils.clip_grad_norm_, (max_norm, norm_type)
    raise RuntimeError(
        f"Incorrect clip grad name ({name}). The valid values are ``clip_grad_value`` "
        "and ``clip_grad_norm``"
    )



from __future__ import annotations

__all__ = ["VanillaOptimizerCreator"]

import logging

from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.optimizers import setup_optimizer

logger = logging.getLogger(__name__)


class VanillaOptimizerCreator(BaseOptimizerCreator):
    r"""Implements a vanilla optimizer creator.

    Args:
    ----
        optimizer_config (dict or ``None``, optional): Specifies the
            optimizer configuration. If ``None``, no optimizer is
            created and ``None`` will be returned by the ``create``
            method. Default: ``None``
        add_module_to_engine (bool, optional): If ``True``, the
            optimizer is added to the engine state, so the optimizer
            state is stored when the engine creates a checkpoint.
            Default: ``True``
    """

    def __init__(
        self, optimizer_config: dict | None = None, add_module_to_engine: bool = True
    ) -> None:
        self._optimizer_config = optimizer_config
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(add_module_to_engine={self._add_module_to_engine})"

    def create(self, engine: BaseEngine, model: Module) -> Optimizer | None:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the optimizer.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
        -------
            ``torch.optim.Optimizer`` or ``None``: The created
                optimizer or ``None`` if there is no optimizer to
                create.
        """
        optimizer = setup_optimizer(model=model, optimizer=self._optimizer_config)
        logger.info(f"optimizer:\n{optimizer}")
        if self._add_module_to_engine and optimizer is not None:
            logger.info(f"Adding an optimizer to the engine (key: {ct.OPTIMIZER})...")
            engine.add_module(ct.OPTIMIZER, optimizer)
        return optimizer



from __future__ import annotations

__all__ = [
    "clone_datapipe",
    "create_chained_datapipe",
    "is_datapipe_config",
    "is_iter_datapipe_config",
    "is_map_datapipe_config",
    "setup_datapipe",
    "setup_iter_datapipe",
    "setup_map_datapipe",
]

from gravitorch.datapipes.factory import (
    create_chained_datapipe,
    is_datapipe_config,
    setup_datapipe,
)
from gravitorch.datapipes.iter import is_iter_datapipe_config, setup_iter_datapipe
from gravitorch.datapipes.map import is_map_datapipe_config, setup_map_datapipe
from gravitorch.datapipes.utils import clone_datapipe



from __future__ import annotations

__all__ = ["DataCreator"]

import copy
import logging
from typing import TypeVar

from gravitorch.data.datacreators.base import BaseDataCreator
from gravitorch.engines.base import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class DataCreator(BaseDataCreator[T]):
    r"""Implement a simple data creator.

    Args:
    ----
        data: Specifies the data to create.
        deepcopy (bool, optional): If ``True``, the data is deep-copied
            before to iterate over the data. It allows a deterministic
            behavior when in-place operations are performed on the
            data. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datacreators import DataCreator
        >>> creator = DataCreator([1, 2, 3, 4])
        >>> creator.create()
        [1, 2, 3, 4]
    """

    def __init__(self, data: T, deepcopy: bool = False) -> None:
        self._data = data
        self._deepcopy = bool(deepcopy)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(deepcopy={self._deepcopy})"

    def create(self, engine: BaseEngine | None = None) -> T:
        data = self._data
        if self._deepcopy:
            try:
                data = copy.deepcopy(data)
            except TypeError:
                logger.warning(
                    "The data can not be deepcopied, please be aware of in-place "
                    "modification would affect the source data."
                )
        return data



__all__ = ["BaseDataFlow", "DataLoaderDataFlow", "IterableDataFlow"]

from gravitorch.experimental.dataflow.base import BaseDataFlow
from gravitorch.experimental.dataflow.dataloader import DataLoaderDataFlow
from gravitorch.experimental.dataflow.iterable import IterableDataFlow



from __future__ import annotations

__all__ = ["create_sequential"]

from collections.abc import Sequence

from torch.nn import Module, Sequential

from gravitorch.nn.utils.factory import setup_module


def create_sequential(modules: Sequence[Module | dict]) -> Sequential:
    r"""Creates a ``torch.nn.Sequential`` from a sequence of modules.

    Args:
    ----
        modules (sequence): Specifies the sequence of modules or their
            configuration.

    Returns:
    -------
        ``torch.nn.Sequential``: The instantiated
            ``torch.nn.Sequential`` module.
    """
    return Sequential(*[setup_module(module) for module in modules])



r"""This package contains the implementation of some profilers."""

__all__ = [
    "BaseProfiler",
    "NoOpProfiler",
    "PyTorchProfiler",
    "is_profiler_config",
    "setup_profiler",
]

from gravitorch.utils.profilers.base import BaseProfiler
from gravitorch.utils.profilers.factory import is_profiler_config, setup_profiler
from gravitorch.utils.profilers.noop import NoOpProfiler
from gravitorch.utils.profilers.pytorch import PyTorchProfiler



from __future__ import annotations

__all__ = ["clone_datapipe"]

import copy
import logging

from torch.utils.data import IterDataPipe, MapDataPipe

logger = logging.getLogger(__name__)


def clone_datapipe(
    datapipe: IterDataPipe | MapDataPipe, raise_error: bool = True
) -> IterDataPipe | MapDataPipe:
    r"""Clone (a.k.a. deepcopy) the ``DataPipe``.

    Args:
    ----
        datapipe (``IterDataPipe`` or ``MapDataPipe``): Specifies the
            ``DataPipe`` to clone.
        raise_error (bool, optional): If ``True``, this function will
            raise an error if the ``DataPipe`` cannot be cloned.
            If ``False``, the original ``DataPipe`` is returned if the
            ``DataPipe`` cannot be cloned. Default: ``True``

    Returns:
    -------
        ``IterDataPipe`` or ``MapDataPipe``: The cloned ``DataPipe``.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import clone_datapipe
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> clone_datapipe(IterableWrapper([1, 2, 1, 3]))
        IterableWrapperIterDataPipe
    """
    try:
        datapipe = copy.deepcopy(datapipe)
    except TypeError as exc:
        if raise_error:
            raise TypeError("The DataPipe can not be cloned") from exc
        logger.warning(
            "The DataPipe can not be cloned, please be aware of in-place "
            "modification would affect source DataPipe."
        )
    return datapipe



from __future__ import annotations

__all__ = [
    "TensorDictShufflerIterDataPipe",
    "shuffle_tensors",
    "shuffle_tensor_mapping",
    "get_first_dimension",
]

import logging
from collections.abc import Iterator, Mapping, Sequence

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_indent
from gravitorch.utils.seed import get_torch_generator
from gravitorch.utils.tensor import permute_along_dim

logger = logging.getLogger(__name__)


class TensorDictShufflerIterDataPipe(IterDataPipe[dict]):
    r"""Implements a source DataPipe to shuffle ``torch.Tensor``s in
    dictionaries.

    For each dictionary, the ``torch.Tensor``s are shuffled with the
    same permutation on a given dimension. This function shuffles
    each tensor on one dimension. It is possible to control the
    dimension used to shuffle the tensors with the argument ``dim``.
    There are two approaches:

        - If ``dim`` is an integer, all the tensors are shuffled on the
            dimension indicated by ``dim``. All the tensors should have
            the same dimension for the specified dimension
            ``(..., dimension, *)`` where ``*`` means any number of
            dimensions.
        - If ``dim`` is a dictionary, each tensor can be shuffled on a
            specific dimension. ``dim`` should indicate the dimension
            to use for each tensor. The keys in this dictionary should
            include keys that are in the input ``mapping``. Note that
            ``dim`` can contain only a subset of keys in the input
            ``mapping``. See example in ``shuffle_tensor_mapping``.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source
            iterable DataPipe.
        dim (int or dict, optional): Specifies the dimension used to
            shuffle the mapping. Default: ``0``
        random_seed (int, optional): Specifies the random seed used
            to shuffle the ``torch.Tensor``s.
            Default: ``3510637111256283951``

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data import IterDataPipe
        >>> from gravitorch.datapipes.iter import TensorDictShuffler
        >>> class MyIterDataPipe(IterDataPipe[dict]):
        ...     def __iter__(self) -> Iterator[dict]:
        ...         for i in range(3):
        ...             yield {"key": torch.arange(4) + i}
        ...
        >>> list(TensorDictShuffler(MyIterDataPipe()))
        [{'key': tensor([2, 3, 0, 1])},
         {'key': tensor([3, 2, 4, 1])},
         {'key': tensor([2, 5, 4, 3])}]
        >>> list(TensorDictShuffler(MyIterDataPipe(), dim={"key": 0}))
        [{'key': tensor([2, 3, 0, 1])},
         {'key': tensor([3, 2, 4, 1])},
         {'key': tensor([2, 5, 4, 3])}]
    """

    def __init__(
        self,
        datapipe: IterDataPipe[Mapping],
        dim: int | dict = 0,
        random_seed: int = 3510637111256283951,
    ) -> None:
        self._datapipe = datapipe
        self._dim = dim
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[dict]:
        for data in self._datapipe:
            yield shuffle_tensor_mapping(data, dim=self._dim, generator=self._generator)

    def __len__(self) -> int:
        return len(self._datapipe)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  dim={self._dim},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe={str_indent(self._datapipe)},\n)"
        )

    @property
    def random_seed(self) -> int:
        r"""``int``: The random seed used to initialize the pseudo
        random generator."""
        return self._generator.initial_seed()


def shuffle_tensors(
    tensors: Sequence[Tensor],
    dim: int = 0,
    generator: torch.Generator | None = None,
) -> list[Tensor]:
    r"""Shuffles tensors with the same permutation on a given dimension.

    This function shuffles the tensors only on a dimension. It is
    possible to control the dimension used to shuffle the tensors with
    the argument ``dim``. All the tensors should have the same
    dimension for the specified dimension ``(..., dimension, *)`` where
    ``*`` means any number of dimensions.

    Args:
    ----
        tensors (``Sequence``): Specifies the tensors to shuffle.
        dim (int, optional): Specifies the dimension used to shuffle
            the tensors. Default: ``0``
        generator (``torch.Generator`` or ``None``, optional):
            Specifies an optional random generator. Default: ``None``

    Returns:
    -------
        ``list``: The shuffled tensors.

    Raises:
    ------
        ValueError if the tensors do not have the same shape for the
            common dimension.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes.iter.shuffling import shuffle_tensors
        >>> shuffle_tensors([torch.arange(4), torch.arange(20).view(4, 5)])
        [
            torch.tensor([0, 2, 1, 3]),
            torch.tensor([[0, 1, 2, 3, 4],
                          [10, 11, 12, 13, 14],
                          [5, 6, 7, 8, 9],
                          [15, 16, 17, 18, 19]]),
        ]
        >>> shuffle_tensors([torch.arange(4).view(1, 4), torch.arange(20).view(5, 4)], dim=1)
        [
            torch.tensor([[0, 2, 1, 3]]),
            torch.tensor([[0, 2, 1, 3],
                          [4, 6, 5, 7],
                          [8, 10, 9, 11],
                          [12, 14, 13, 15],
                          [16, 18, 17, 19]]),
        ]
    """
    dims = [tensor.shape[dim] for tensor in tensors]
    if len(set(dims)) > 1:
        raise ValueError(
            f"The tensors do not have the same shape for the common dimension (dim={dim}): {dims}"
        )
    permutation = torch.randperm(tensors[0].shape[dim], generator=generator)
    return [permute_along_dim(tensor, permutation, dim) for tensor in tensors]


def shuffle_tensor_mapping(
    mapping: Mapping, dim: int | dict = 0, generator: torch.Generator | None = None
) -> dict:
    r"""Shuffles the tensors that are in a mapping with the same
    permutation.

    This function shuffles each tensor on one dimension. It is
    possible to control the dimension used to shuffle the tensors
    with the argument ``dim``. There are two approaches:

        - If ``dim`` is an integer, all the tensors are shuffled on the
            dimension indicated by ``dim``. All the tensors should have
            the same dimension for the specified dimension
            ``(..., dimension, *)`` where ``*`` means any number of
            dimensions.
        - If ``dim`` is a dictionary, each tensor can be shuffled on a
            specific dimension. ``dim`` should indicate the dimension
            to use for each tensor. The keys in this dictionary should
            include keys that are in the input ``mapping``. Note that
            ``dim`` can contain only a subset of keys in the input
            ``mapping``. See example below.

    Args:
    ----
        mapping (``Mapping``): Specifies the mapping with the tensors
            to shuffle.
        dim (int or dict, optional): Specifies the dimension used to
            shuffle the mapping. Default: ``0``
        generator (``torch.Generator`` or ``None``, optional):
            Specifies an optional random generator. Default: ``None``

    Returns:
    -------
        ``dict``: A dictionary with the shuffled tensors.

    Raises:
    ------
        ValueError if the tensors do not have the same shape for the
            common dimension.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes.iter.shuffling import shuffle_tensor_mapping
        # Shuffle all the tensors on dimension 0 (default)
        >>> shuffle_tensor_mapping({"key1": torch.arange(4), "key2": torch.arange(20).view(4, 5)})
        {
            "key1": tensor([0, 2, 1, 3]),
            "key2": tensor([[0, 1, 2, 3, 4],
                            [10, 11, 12, 13, 14],
                            [5, 6, 7, 8, 9],
                            [15, 16, 17, 18, 19]]),
        }
        # Shuffle all the tensors on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4).view(1, 4), "key2": torch.arange(20).view(5, 4)},
        ...     dim=1,
        ... )
        {
            "key1": tensor([[0, 2, 1, 3]]),
            "key2": tensor([[0, 2, 1, 3],
                            [4, 6, 5, 7],
                            [8, 10, 9, 11],
                            [12, 14, 13, 15],
                            [16, 18, 17, 19]]),
        }
        # Shuffle the tensor "key1" on dimension 0 and the tensor "key2" on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4), "key2": torch.arange(20).view(5, 4)},
        ...     dim={"key1": 0, "key2": 1},
        ... )
        {
            'key1': tensor([1, 2, 0, 3]),
            'key2': tensor([[1,  2,  0,  3],
                            [5,  6,  4,  7],
                            [9, 10,  8, 11],
                            [13, 14, 12, 15],
                            [17, 18, 16, 19]]),
        }
        # Shuffle only the tensor "key2" on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4), "key2": torch.arange(20).view(5, 4)},
        ...     dim={"key2": 1},
        ... )
        {
            'key1': tensor([0, 1, 2, 3]),
            'key2': tensor([[3,  1,  0,  2],
                            [7,  5,  4,  6],
                            [11,  9,  8, 10],
                            [15, 13, 12, 14],
                            [19, 17, 16, 18]]),
        }
    """
    dims = dim if isinstance(dim, dict) else {key: dim for key in mapping}
    if len(dims) == 0:  # No tensor to permute
        return {key: tensor for key, tensor in mapping.items()}
        # Check if the common dimensions are the same
    valid_dims = [mapping[key].shape[dims[key]] for key in mapping if key in dims]
    if len(set(valid_dims)) > 1:
        raise ValueError(
            f"The tensors do not have the same shape for the common dimension: {valid_dims}"
        )
    # Permute the tensors
    key = next(iter(dims.keys()))
    permutation = torch.randperm(mapping[key].shape[dims[key]], generator=generator)
    output = {}
    for key, tensor in mapping.items():
        if key in dims:
            tensor = permute_along_dim(tensor, permutation, dims[key])
        output[key] = tensor
    return output


def get_first_dimension(data: Tensor | np.ndarray | list | tuple) -> int:
    r"""Finds the first dimension value.

    This function finds the first dimension value by using the
    following rules:

        - if the input is a ``torch.Tensor`` or ``numpy.ndarray``,
            the first dimension value is returned.
        - if the input is a list or tuple, the length is returned.

    Args:
    ----
        data: Specifies the data to get the first dimension value.

    Returns:
    -------
        int: The first dimension value.

    Raises:
    ------
        TypeError: if the input type is not supported.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes.iter.shuffling import get_first_dimension
        # torch tensor
        >>> import torch
        >>> get_first_dimension(torch.ones(5, 4))
        5
        # NumPy ndarray
        >>> import numpy
        >>> get_first_dimension(numpy.ones((7, 4)))
        7
        # List/tuple
        >>> get_first_dimension([1, 3, 4])
        3
        >>> get_first_dimension((5, 2, 1, 4))
        4
    """
    if torch.is_tensor(data) or isinstance(data, np.ndarray):
        return data.shape[0]
    if isinstance(data, (list, tuple)):
        return len(data)
    raise TypeError(
        f"The supported types are: torch.Tensor, numpy.ndarray, list and tuple. (received: {data})"
    )



r"""This module defines the base class to implement training loops."""
__all__ = ["BaseTrainingLoop"]

import logging
from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine

logger = logging.getLogger(__name__)


class BaseTrainingLoop(ABC, metaclass=AbstractFactory):
    r"""Defines the base class ti implement training loops.

    To implement your own training loop, you will need to define the
    following methods:

        - ``train``
        - ``load_state_dict``
        - ``state_dict``
    """

    @abstractmethod
    def train(self, engine: BaseEngine) -> None:
        r"""Trains the model on the training dataset.

        The training metrics/artifacts should be logged through the
        engine.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
        """

    @abstractmethod
    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Sets up the training loop from a dictionary containing the
        state values.

        Args:
        ----
            state_dict (dict): Specifies a dictionary
                containing state keys with values.
        """

    @abstractmethod
    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns
        -------
            dict: The state values in a dict.
        """



r"""This module contains some tools to analyze the parameters of a
``torch.nn.Module``."""

from __future__ import annotations

__all__ = [
    "ParameterSummary",
    "get_parameter_summaries",
    "show_parameter_summary",
]

import logging
from dataclasses import asdict, dataclass

import torch
from tabulate import tabulate
from torch.nn import Module, Parameter, UninitializedParameter

from gravitorch.utils.mapping import convert_to_dict_of_lists

logger = logging.getLogger(__name__)


@dataclass
class ParameterSummary:
    r"""Implements a class to easily manage parameter summaries.

    NI: Not Initialized
    NP: No Parameter
    """
    name: str
    mean: float | str
    median: float | str
    std: float | str
    min: float | str
    max: float | str
    shape: tuple[int, ...] | str
    learnable: bool | str
    device: torch.device | str

    @classmethod
    def from_parameter(
        cls, name: str, parameter: Parameter | UninitializedParameter
    ) -> ParameterSummary:
        r"""Creates the parameter summary from the parameter object.

        Args:
            name (str): Specifies the name of the parameter.
            parameter (``torch.nn.Parameter`` or
                ``torch.nn.UninitializedParameter``): Specifies the
                parameter object.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from torch.nn import Parameter
            >>> from gravitorch.nn import ParameterSummary
            >>> ParameterSummary.from_parameter("weight", Parameter(torch.randn(6, 4)))
            ParameterSummary(name='weight', mean=0.10939589142799377, median=-0.018833406269550323,
            std=0.8310111165046692, min=-1.6590238809585571, max=2.0370750427246094,
            learnable=True, shape=(6, 4))
        """
        if isinstance(parameter, UninitializedParameter):
            return cls(
                name=name,
                mean="NI",
                median="NI",
                std="NI",
                min="NI",
                max="NI",
                shape="NI",
                learnable=parameter.requires_grad,
                device=parameter.device,
            )
        if parameter.numel() == 0:
            return cls(
                name=name,
                mean="NP",
                median="NP",
                std="NP",
                min="NP",
                max="NP",
                shape=tuple(parameter.shape),
                learnable=parameter.requires_grad,
                device=parameter.device,
            )
        return cls(
            name=name,
            mean=parameter.mean().item(),
            median=parameter.median().item(),
            std=parameter.std(dim=None).item(),
            min=parameter.min().item(),
            max=parameter.max().item(),
            shape=tuple(parameter.shape),
            learnable=parameter.requires_grad,
            device=parameter.device,
        )


def get_parameter_summaries(module: Module) -> list[ParameterSummary]:
    r"""Gets the parameter summaries of a module.

    Args:
        module (``torch.nn.Module``): Specifies the module with the
            parameters to summarize.

    Returns:
        list: The list of parameter summaries.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from torch.nn import Linear
        >>> from gravitorch.nn import get_parameter_summaries
        >>> get_parameter_summaries(Linear(4, 6))
        [ParameterSummary(name='weight', mean=0.10450785607099533, median=0.02029263973236084,
          std=0.26641708612442017, min=-0.4861736297607422, max=0.48399144411087036,
          learnable=True, shape=(6, 4)),
         ParameterSummary(name='bias', mean=0.11823087930679321, median=-0.05595135688781738,
          std=0.36556652188301086, min=-0.4497765898704529, max=0.4593251347541809,
          learnable=True, shape=(6,))]
    """
    return [
        ParameterSummary.from_parameter(name, parameter)
        for name, parameter in module.named_parameters()
    ]


def show_parameter_summary(
    module: Module, tablefmt: str = "fancy_outline", floatfmt: str = ".6f"
) -> None:
    r"""Shows a summary of the model parameters.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to analyze.
        tablefmt (str, optional): Specifies the table format.
            Default: ``'fancy_outline'``
        floatfmt (str, optional): Specifies the float format.
            Default: ``'.6f'``
    """
    summaries = convert_to_dict_of_lists(
        [asdict(summary) for summary in get_parameter_summaries(module)]
    )
    logger.info(
        "Parameter summary\n"
        f'{tabulate(summaries, headers="keys", tablefmt=tablefmt, floatfmt=floatfmt)}\n'
    )



__all__ = [
    "BaseInitializer",
    "Constant",
    "ConstantBias",
    "KaimingNormal",
    "KaimingUniform",
    "NoOpInitializer",
    "SequentialInitializer",
    "TruncNormal",
    "XavierNormal",
    "XavierUniform",
    "constant",
    "constant_bias",
    "kaiming_normal",
    "kaiming_uniform",
    "setup_initializer",
    "trunc_normal",
    "xavier_normal",
    "xavier_uniform",
]

from gravitorch.nn.init.base import BaseInitializer
from gravitorch.nn.init.const import Constant, ConstantBias, constant, constant_bias
from gravitorch.nn.init.factory import setup_initializer
from gravitorch.nn.init.kaiming import (
    KaimingNormal,
    KaimingUniform,
    kaiming_normal,
    kaiming_uniform,
)
from gravitorch.nn.init.noop import NoOpInitializer
from gravitorch.nn.init.normal import TruncNormal, trunc_normal
from gravitorch.nn.init.sequential import SequentialInitializer
from gravitorch.nn.init.xavier import (
    XavierNormal,
    XavierUniform,
    xavier_normal,
    xavier_uniform,
)



r"""The data package contains the data loader base class and some tools
to speed up the implementation or setup of new data loaders."""

from __future__ import annotations

__all__ = [
    "create_dataloader",
    "create_dataloader2",
    "is_dataloader_config",
    "is_dataloader2_config",
    "setup_dataloader",
    "setup_dataloader2",
]

from gravitorch.data.dataloaders.factory import (
    create_dataloader,
    create_dataloader2,
    is_dataloader2_config,
    is_dataloader_config,
    setup_dataloader,
    setup_dataloader2,
)



from __future__ import annotations

__all__ = [
    "BaseDataSource",
    "DataCreatorIterDataPipeCreatorDataSource",
    "DatasetDataSource",
    "ImageNetDataSource",
    "IterDataPipeCreatorDataSource",
    "LoaderNotFoundError",
    "is_datasource_config",
    "setup_and_attach_datasource",
    "setup_datasource",
]

from gravitorch.datasources.base import (
    BaseDataSource,
    LoaderNotFoundError,
    is_datasource_config,
    setup_and_attach_datasource,
    setup_datasource,
)
from gravitorch.datasources.datapipe import (
    DataCreatorIterDataPipeCreatorDataSource,
    IterDataPipeCreatorDataSource,
)
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.datasources.imagenet import ImageNetDataSource



from __future__ import annotations

__all__ = ["create_chained_datapipe", "is_datapipe_config", "setup_datapipe"]

import logging
from collections.abc import Sequence
from typing import TypeVar

from objectory import OBJECT_TARGET, factory
from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.datapipes.iter import is_iter_datapipe_config
from gravitorch.datapipes.map import is_map_datapipe_config
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


def create_chained_datapipe(
    config: dict | Sequence[dict],
    source_inputs: Sequence | None = None,
) -> IterDataPipe[T] | MapDataPipe[T]:
    r"""Creates a chained ``DataPipe`` object.

    A chained ``DataPipe`` object has a single source (which
    can takes multiple ``DataPipe`` objects) and a single sink.
    The structure should look like:

        SourceDatapipe -> DataPipe1 -> DataPipe2 -> SinkDataPipe

    The structure of the ``config`` input depends on the chained
    ``DataPipe`` object that is created:

        - If ``config`` is a ``dict`` object, it creates a chained
            ``DataPipe`` object with a single ``DataPipe``
            object. The dictionary should contain the parameters used
            to initialize the ``DataPipe`` object. It should
            follow the ``object_factory`` syntax. Using a dict allows
            to initialize a single ``DataPipe`` object. If you
            want to create a ``DataPipe`` object recursively, you
            need to give a sequence of dict.
        - If ``config`` is a sequence of ``dict`` objects, this
            function creates an ``DataPipe`` object with a
            chained structure. The sequence of configurations
            follows the order of the ``DataPipe``s. The first
            config is used to create the first ``DataPipe``
            (a.k.a. source), and the last config is used to create the
            last ``DataPipe`` (a.k.a. sink). This function assumes
            all the DataPipes have a single source DataPipe as their
            first argument, excepts for the source ``DataPipe``.

    Note: it is possible to create chained ``DataPipe`` objects
    without using this function.

    Args:
    ----
        config (dict or sequence of dict): Specifies the configuration
            of the ``DataPipe`` object to create. See description
            above to know when to use a dict or a sequence of dicts.
        source_inputs (sequence or ``None``): Specifies the first
            positional arguments of the source ``DataPipe``. This
            argument can be used to create a new ``DataPipe``
            object, that takes existing ``DataPipe`` objects as
            input. See examples below to see how to use it.
            If ``None``, ``source_inputs`` is set to an empty tuple.
            Default: ``None``

    Returns:
    -------
        ``DataPipe``: The last (a.k.a. sink) ``DataPipe`` of
            the sequence.

    Raises:
    ------
        RuntimeError if the configuration is empty (empty dict or
            sequence).

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import create_chained_datapipe
        # Create an DataPipe object using a single DataPipe object and no source input
        >>> datapipe = create_chained_datapipe(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # Equivalent to
        >>> datapipe = create_chained_datapipe(
        ...     [
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4],
        ...         },
        ...     ]
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # It is possible to use the source_inputs to create the same DataPipe object.
        # The data is given by the source_inputs
        >>> datapipe = create_chained_datapipe(
        ...     config={"_target_": "torch.utils.data.datapipes.iter.IterableWrapper"},
        ...     source_inputs=([1, 2, 3, 4],),
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # Create an DataPipe object using two DataPipe objects and no source input
        >>> datapipe = create_chained_datapipe(
        ...     [
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4],
        ...         },
        ...         {"_target_": "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ]
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        # It is possible to use the source_inputs to create the same DataPipe object.
        # A source DataPipe object is specified by using source_inputs
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> datapipe = create_chained_datapipe(
        ...     config=[
        ...         {"_target_": "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ],
        ...     source_inputs=[IterableWrapper(data=[1, 2, 3, 4])],
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        # It is possible to create a chained ``DataPipe`` object that takes several
        # DataPipe objects as input.
        >>> datapipe = create_chained_datapipe(
        ...     config=[
        ...         {"_target_": "torch.utils.data.datapipes.iter.Multiplexer"},
        ...         {"_target_": "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ],
        ...     source_inputs=[
        ...         IterableWrapper(data=[1, 2, 3, 4]),
        ...         IterableWrapper(data=[11, 12, 13, 14]),
        ...     ],
        ... )
        >>> tuple(datapipe)
        ([1, 11], [2, 12], [3, 13], [4, 14])
    """
    if not config:
        raise RuntimeError("It is not possible to create a DataPipe because the config is empty")
    source_inputs = source_inputs or ()
    if isinstance(config, dict):
        config = config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        return factory(target, *source_inputs, **config)
    datapipe = create_chained_datapipe(config[0], source_inputs)
    for cfg in config[1:]:
        datapipe = create_chained_datapipe(cfg, source_inputs=(datapipe,))
    return datapipe


def is_datapipe_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``DataPipe``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``DataPipe`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import is_datapipe_config
        >>> is_datapipe_config(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        True
        >>> is_datapipe_config(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.map.SequenceWrapper",
        ...         "sequence": [1, 2, 3, 4],
        ...     }
        ... )
        True
    """
    return is_iter_datapipe_config(config) or is_map_datapipe_config(config)


def setup_datapipe(
    datapipe: IterDataPipe[T] | MapDataPipe[T] | dict,
) -> IterDataPipe[T] | MapDataPipe[T]:
    r"""Sets up a ``torch.utils.data.graph.DataPipe`` object.

    Args:
    ----
        datapipe (``IterDataPipe`` or ``MapDataPipe`` or dict):
            Specifies the datapipe or its configuration (dictionary).

    Returns:
    -------
        ``IterDataPipe`` or ``MapDataPipe``: The instantiated
            ``DataPipe`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import setup_datapipe
        >>> datapipe = setup_datapipe(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        >>> datapipe
        IterableWrapperIterDataPipe
        >>> setup_datapipe(datapipe)  # Do nothing because the datapipe is already instantiated
        IterableWrapperIterDataPipe
    """
    if isinstance(datapipe, dict):
        logger.info(
            "Initializing a `torch.utils.data.graph.DataPipe` from its configuration... "
            f"{str_target_object(datapipe)}"
        )
        datapipe = factory(**datapipe)
    return datapipe



r"""This module implements some utility functions for the profilers."""

from __future__ import annotations

__all__ = ["is_profiler_config", "setup_profiler"]

import logging

from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object
from gravitorch.utils.profilers.base import BaseProfiler
from gravitorch.utils.profilers.noop import NoOpProfiler

logger = logging.getLogger(__name__)


def is_profiler_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseProfiler``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseProfiler`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.profilers import is_profiler_config
        >>> is_profiler_config({"_target_": "gravitorch.utils.profilers.NoOpProfiler"})
        True
    """
    return is_object_config(config, BaseProfiler)


def setup_profiler(profiler: BaseProfiler | dict | None) -> BaseProfiler:
    r"""Sets up the profiler.

    The profiler is instantiated from its configuration by using the
    ``BaseProfiler`` factory function.

    Args:
    ----
        profiler (``BaseProfiler`` or dict or None): Specifies the
            profiler or its configuration. If ``None``, the
            ``NoOpProfiler`` is instantiated.

    Returns:
    -------
        ``BaseProfiler``: A profiler.
    """
    if profiler is None:
        profiler = NoOpProfiler()
    if isinstance(profiler, dict):
        logger.info(
            f"Initializing a profiler from its configuration... {str_target_object(profiler)}"
        )
        profiler = BaseProfiler.factory(**profiler)
    return profiler



__all__ = [
    "AMPTrainingLoop",
    "AccelerateTrainingLoop",
    "BaseBasicTrainingLoop",
    "BaseTrainingLoop",
    "NoOpTrainingLoop",
    "VanillaTrainingLoop",
    "is_training_loop_config",
    "setup_training_loop",
]

from gravitorch.loops.training.accelerate import AccelerateTrainingLoop
from gravitorch.loops.training.amp import AMPTrainingLoop
from gravitorch.loops.training.base import BaseTrainingLoop
from gravitorch.loops.training.basic import BaseBasicTrainingLoop
from gravitorch.loops.training.factory import (
    is_training_loop_config,
    setup_training_loop,
)
from gravitorch.loops.training.noop import NoOpTrainingLoop
from gravitorch.loops.training.vanilla import VanillaTrainingLoop



r"""This module implements some utility functions for the evaluation
loops."""

__all__ = ["is_evaluation_loop_config", "setup_evaluation_loop"]

import logging
from typing import Union

from objectory.utils import is_object_config

from gravitorch.loops.evaluation.base import BaseEvaluationLoop
from gravitorch.loops.evaluation.vanilla import VanillaEvaluationLoop
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_evaluation_loop_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseEvaluationLoop``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseEvaluationLoop`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.evaluation import is_evaluation_loop_config
        >>> is_evaluation_loop_config(
        ...     {"_target_": "gravitorch.loops.evaluation.VanillaEvaluationLoop"}
        ... )
        True
    """
    return is_object_config(config, BaseEvaluationLoop)


def setup_evaluation_loop(
    evaluation_loop: Union[BaseEvaluationLoop, dict, None]
) -> BaseEvaluationLoop:
    r"""Sets up the evaluation loop.

    The evaluation loop is instantiated from its configuration by
    using the ``BaseEvaluationLoop`` factory function.

    Args:
    ----
        evaluation_loop (``BaseEvaluationLoop`` or dict or None):
            Specifies the evaluation loop or its configuration.
            If ``None``, the ``VanillaEvaluationLoop`` is instantiated.

    Returns:
    -------
        ``BaseEvaluationLoop``: The evaluation loop.
    """
    if evaluation_loop is None:
        evaluation_loop = VanillaEvaluationLoop()
    if isinstance(evaluation_loop, dict):
        logger.info(
            f"Initializing an evaluation loop from its configuration... "
            f"{str_target_object(evaluation_loop)}"
        )
        evaluation_loop = BaseEvaluationLoop.factory(**evaluation_loop)
    if not isinstance(evaluation_loop, BaseEvaluationLoop):
        logger.warning(
            f"evaluation_loop is not a `BaseEvaluationLoop` (received: {type(evaluation_loop)})"
        )
    return evaluation_loop



__all__ = ["is_loop_observer_config", "setup_loop_observer"]

import logging
from typing import Union

from objectory.utils import is_object_config

from gravitorch.loops.observers.base import BaseLoopObserver
from gravitorch.loops.observers.noop import NoOpLoopObserver
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_loop_observer_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseLoopObserver``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseLoopObserver`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.observers import is_loop_observer_config
        >>> is_loop_observer_config({"_target_": "gravitorch.loops.observers.NoOpLoopObserver"})
        True
    """
    return is_object_config(config, BaseLoopObserver)


def setup_loop_observer(loop_observer: Union[BaseLoopObserver, dict, None]) -> BaseLoopObserver:
    r"""Sets up a loop observer.

    The loop observer is instantiated from its configuration by
    using the ``BaseLoopObserver`` factory function.

    Args:
    ----
        loop_observer (``BaseLoopObserver`` or dict or None):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.

    Returns:
    -------
        ``BaseLoopObserver``: The loop observer.
    """
    if loop_observer is None:
        loop_observer = NoOpLoopObserver()
    if isinstance(loop_observer, dict):
        logger.info(
            "Initializing a loop observer module from its configuration... "
            f"{str_target_object(loop_observer)}"
        )
        loop_observer = BaseLoopObserver.factory(**loop_observer)
    if not isinstance(loop_observer, BaseLoopObserver):
        logger.warning(
            f"loop_observer is not a `BaseLoopObserver` (received: {type(loop_observer)})"
        )
    return loop_observer



from __future__ import annotations

__all__ = ["create_sequential_iter_datapipe", "is_iter_datapipe_config", "setup_iter_datapipe"]

import logging
from collections.abc import Sequence

from objectory import OBJECT_TARGET, factory
from objectory.utils import is_object_config
from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def create_sequential_iter_datapipe(configs: Sequence[dict]) -> IterDataPipe:
    r"""Creates a sequence of ``IterDataPipe``s from their configuration.

    Args:
    ----
        configs: Specifies the config of each ``IterDataPipe`` in the
            sequence. The configs sequence follows the order of the
            ``IterDataPipe``s. The first config is used to create the
            first ``IterDataPipe`` (a.k.a. source), and the last
            config is used to create the last ``IterDataPipe``
            (a.k.a. sink). This function assumes all the DataPipes
            have a single source DataPipe as their first argument,
            excepts for the first DataPipe.

    Returns:
    -------
        ``IterDataPipe``: The last ``IterDataPipe`` of the sequence
            (a.k.a. sink).

    Raises:
    ------
        ValueError if the ``IterDataPipe`` configuration sequence is
            empty.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes.iter.factory import create_sequential_iter_datapipe
        >>> datapipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             "_target_": "gravitorch.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...     ],
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> datapipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             "_target_": "gravitorch.datapipes.iter.SourceWrapper",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...         {"_target_": "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ]
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
    """
    if not configs:
        raise ValueError("It is not possible to create a DataPipe because the configs are empty")
    datapipe = factory(**configs[0])
    for config in configs[1:]:
        config = config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        datapipe = factory(target, datapipe, **config)
    return datapipe


def is_iter_datapipe_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``IterDataPipe``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``IterDataPipe`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import is_iter_datapipe_config
        >>> is_iter_datapipe_config(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        True
    """
    return is_object_config(config, IterDataPipe)


def setup_iter_datapipe(datapipe: IterDataPipe | Sequence[dict]) -> IterDataPipe:
    r"""Sets up an ``IterDataPipe``.

    Note: it is only possible to create a sequential DataPipe from
    its configuration. See ``create_sequential_iter_datapipe``
    documentation for more information.

    Args:
    ----
        datapipe: Specifies the DataPipe or its configuration.

    Returns:
    -------
        ``IterDataPipe``: The last ``IterDataPipe`` of the sequence
            (a.k.a. sink).

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import setup_iter_datapipe
        >>> datapipe = setup_iter_datapipe(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     },
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
    """
    if isinstance(datapipe, dict):
        logger.info(
            f"Initializing a `IterDataPipe` from its configuration... {str_target_object(datapipe)}"
        )
        datapipe = factory(**datapipe)
    return datapipe



__all__ = ["EarlyStopping"]

import logging
import operator
from collections.abc import Callable

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.events import VanillaEventHandler
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.history import BaseHistory, MaxScalarHistory, MinScalarHistory

logger = logging.getLogger(__name__)


class EarlyStopping(BaseHandler):
    r"""Implements an early stopping handler to stop the training if no
    improvement after a given number of epochs.

    This early stopping handler only works with
    ``MaxScalarHistory`` and ``MinScalarHistory``
    metrics.

    Args:
    ----
        metric_name (str, optional): Specifies the metric name that
            is used to measure the improvement. By default, the metric
            is the loss on the evaluation data set.
            Default: 'eval/loss'.
        patience (int, optional): Specifies the number of epochs to
            wait if no improvement and then stop the training.
            Default: ``5``
        delta (float, optional): A minimum increase or decrease in
            the score to qualify as an improvement, i.e. an increase
            of less than or equal to `delta`, will count as no
            improvement. Default: ``0.0``
        cumulative_delta (bool, optional): It True, `min_delta`
            defines an increase since the last `patience` reset,
            otherwise, it defines an increase after the last epoch.
            Default: ``False``.

    Example usage:

    .. code-block:: pycon

        # Create an engine
        >>> engine = ...
        >>> from gravitorch.handlers import EarlyStopping
        # Add the early stopping handler to the engine with a patience of 10 epochs
        # for the accuracy metric.
        >>> handler = EarlyStopping(metric_name="eval/accuracy", patience=10)
        >>> handler.attach(engine)
    """

    def __init__(
        self,
        metric_name: str = f"{ct.EVAL}/loss",
        patience: int = 5,
        delta: float = 0.0,
        cumulative_delta: bool = False,
    ) -> None:
        self._metric_name = str(metric_name)
        if patience < 1:
            raise ValueError(f"patience must be a positive integer (received: {patience})")
        self._patience = int(patience)
        if delta < 0.0:
            raise ValueError(f"delta should not be a negative number (received: {patience})")
        self._delta = float(delta)
        self._cumulative_delta = bool(cumulative_delta)

        self._waiting_counter = 0
        self._best_score = None
        self._best_epoch = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(metric_name={self._metric_name}, "
            f"patience={self._patience}, "
            f"delta={self._delta}, "
            f"cumulative_delta={self._cumulative_delta}, "
            f"waiting_counter={self._waiting_counter}, "
            f"best_score={self._best_score}, "
            f"best_epoch={self._best_epoch})"
        )

    def attach(self, engine: BaseEngine) -> None:
        logger.info(
            f"Attach early stopping handler for metric {self._metric_name} with the parameters: "
            f"patience={self._patience:,}   delta={self._delta}   "
            f"cumulative_delta={self._cumulative_delta}"
        )
        add_unique_event_handler(
            engine=engine,
            event=EngineEvents.TRAIN_STARTED,
            event_handler=VanillaEventHandler(self.start, handler_kwargs={"engine": engine}),
        )
        add_unique_event_handler(
            engine=engine,
            event=EngineEvents.EPOCH_COMPLETED,
            event_handler=VanillaEventHandler(self.step, handler_kwargs={"engine": engine}),
        )

        # Add the module to the engine so the state of the early stopping module is saved when
        # the engine checkpoint is created. It is necessary to store the state to resume
        # the training.
        engine.add_module(ct.EARLY_STOPPING, self)

        if engine.has_history(self._metric_name):
            self._check_history(engine.get_history(self._metric_name))
        else:
            logger.warning(
                f"There is no history for '{self._metric_name}'. The history needs to be created "
                f"before the `step` method is called"
            )

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
        ----
            state_dict (dict): a dict with parameters

        Example:
        -------
        .. code-block:: pycon

            >>> handler = EarlyStopping(...)
            >>> state = {...}
            >>> handler.load_state_dict(state)
        """
        self._best_epoch = state_dict["best_epoch"]
        self._best_score = state_dict["best_score"]
        self._waiting_counter = state_dict["waiting_counter"]

    def state_dict(self) -> dict:
        r"""Gets a dictionary containing state values.

        Returns:
        -------
            dict: the state values in a dict.

        Example:
        -------
        .. code-block:: pycon

            >>> state = EarlyStopping(...)
            >>> state_dict = state.state_dict()
        """
        return {
            "best_epoch": self._best_epoch,
            "best_score": self._best_score,
            "waiting_counter": self._waiting_counter,
        }

    def start(self, engine: BaseEngine) -> None:
        r"""Stops the training if the requirements to stop the training
        are already met.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info(
            f"Early stopping [start] | {self._metric_name} | "
            f"best_epoch {self._best_epoch} | best_score {self._best_score} | "
            f"waiting_counter {self._waiting_counter}/{self._patience}"
        )
        if self._waiting_counter >= self._patience:
            logger.info(
                f"Early stopping [start] | cancel training because '{self._metric_name}' did not "
                f"improve during the last {self._waiting_counter} epochs"
            )
            engine.terminate()

    def step(self, engine: BaseEngine) -> None:
        """Updates the early stopping handler by using the last value of
        the monitored metric.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
        """
        history = engine.get_history(self._metric_name)
        current_score = history.get_last_value()
        min_delta = self._get_min_delta(history)
        comparator = self._get_comparator(history)

        if self._best_score is None:
            # The best score is undefined at the first call.
            # The first score becomes the best score.
            self._best_score = current_score
            self._best_epoch = engine.epoch

        elif not comparator(current_score - min_delta, self._best_score):
            # The current score is not better than the best score + min delta.
            self._waiting_counter += 1
            if not self._cumulative_delta and comparator(current_score, self._best_score):
                self._best_score = current_score
            if self._waiting_counter >= self._patience:
                logger.info(
                    f"Early stopping [step] | stop training because '{self._metric_name}' did not "
                    f"improve during the last {self._waiting_counter:,} epochs"
                )
                engine.terminate()

        else:
            # The current score is better than the best score.
            self._best_score = current_score
            self._waiting_counter = 0
            self._best_epoch = engine.epoch

        logger.info(
            f"Early stopping [step] | {self._metric_name} | "
            f"best_epoch {self._best_epoch} | best_score {self._best_score} | "
            f"waiting_counter {self._waiting_counter}/{self._patience}"
        )

    def _check_history(self, history: BaseHistory) -> None:
        r"""Checks the history.

        Args:
        ----
            history (``BaseHistory``): Specifies the history
                tracker to check.

        Raises:
        ------
            RuntimeError if the history tracker is not valid.
        """
        if not isinstance(history, (MaxScalarHistory, MinScalarHistory)):
            raise RuntimeError(
                f"The early stopping handler only supports ``MaxScalarHistory`` "
                f"or ``MinScalarHistory`` history tracker (received: {history})"
            )

    def _get_comparator(self, history: BaseHistory) -> Callable:
        r"""Gets the comparator function from the history tracker type.

        Args:
        ----
            history (``BaseHistory``): Specifies the history
                tracker.

        Returns:
        -------
            callable: The comparator.
        """
        if isinstance(history, MaxScalarHistory):
            return operator.gt  # >
        return operator.lt  # <

    def _get_min_delta(self, history: BaseHistory) -> float:
        r"""Gets the minimum delta.

        Args:
        ----
            history (``BaseHistory``): Specifies the history
                tracker.

        Returns:
        -------
            float: The minimum delta.
        """
        self._check_history(history)
        return self._delta if isinstance(history, MaxScalarHistory) else -self._delta



from __future__ import annotations

__all__ = ["CompiledModelCreator"]

import logging

import torch
from torch.nn import Module

from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_indent, str_pretty_json

logger = logging.getLogger(__name__)


class CompiledModelCreator(BaseModelCreator):
    r"""Implements a model creator that compiles a model with
    ``torch.compile``.

    Args:
    ----
        model_creator (``BaseModelCreator`` or dict): Specifies a
            model creator or its configuration. The created model
            should be compatible with ``DistributedDataParallel``.
        compile_kwargs (dict or ``None``): Specifies some keyword
            arguments used to compile the model. Please read the
            documentation of ``torch.compile`` to see the possible
            options. Default: ``None``
    """

    def __init__(
        self,
        model_creator: BaseModelCreator | dict,
        compile_kwargs: dict | None = None,
    ) -> None:
        self._model_creator = setup_model_creator(model_creator)
        self._compile_kwargs = compile_kwargs or {}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  model_creator={self._model_creator},\n"
            f"  compile_kwargs={str_indent(str_pretty_json(self._compile_kwargs))},\n"
            ")"
        )

    def create(self, engine: BaseEngine) -> Module:
        return torch.compile(self._model_creator.create(engine), **self._compile_kwargs)



__all__ = ["BaseBatchLoader"]

from collections.abc import Iterable
from contextlib import AbstractContextManager
from typing import TypeVar

T = TypeVar("T")


class BaseBatchLoader(Iterable[T], AbstractContextManager):
    r"""Defines the base class to implement batch loader."""



__all__ = ["BaseBatchLoader", "IterableBatchLoader"]

from gravitorch.experimental.batchloader.base import BaseBatchLoader
from gravitorch.experimental.batchloader.iterable import IterableBatchLoader



from __future__ import annotations

__all__ = ["setup_module"]

import logging

from objectory import factory
from objectory.utils import is_object_config
from torch.nn import Module

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_module_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``torch.nn.Module``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``torch.nn.Module`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn import is_module_config
        >>> is_module_config({"_target_": "torch.nn.Identity"})
        True
    """
    return is_object_config(config, Module)


def setup_module(module: Module | dict) -> Module:
    r"""Sets up a ``torch.nn.Module`` object.

    Args:
    ----
        module (``torch.nn.Module`` or dict): Specifies the module or
            its configuration (dictionary).

    Returns:
    -------
        ``torch.nn.Module``: The instantiated ``torch.nn.Module``
            object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn import setup_module
        >>> linear = setup_module(
        ...     {"_target_": "torch.nn.Linear", "in_features": 4, "out_features": 6}
        ... )
        >>> linear
        Linear(in_features=4, out_features=6, bias=True)
        >>> setup_module(linear)  # Do nothing because the module is already instantiated
        Linear(in_features=4, out_features=6, bias=True)
    """
    if isinstance(module, dict):
        logger.info(
            "Initializing a `torch.nn.Module` from its configuration... "
            f"{str_target_object(module)}"
        )
        module = factory(**module)
    if not isinstance(module, Module):
        logger.warning(f"module is not a `torch.nn.Module` (received: {type(module)})")
    return module



from __future__ import annotations

__all__ = ["DictBatcherIterDataPipe", "TupleBatcherIterDataPipe"]

import logging
from collections.abc import Hashable, Iterator, Sequence

import torch
from coola import summary
from torch import Tensor
from torch.utils.data import IterDataPipe

from gravitorch.datapipes.iter.shuffling import shuffle_tensor_mapping, shuffle_tensors
from gravitorch.datapipes.iter.source import SourceWrapperIterDataPipe
from gravitorch.utils.format import str_indent
from gravitorch.utils.mapping import get_first_value
from gravitorch.utils.seed import get_torch_generator

logger = logging.getLogger(__name__)


class DictBatcherIterDataPipe(IterDataPipe[dict]):
    r"""Implements a source DataPipe to generate batch of examples from a
    dictionary of ``torch.Tensor``s.

    Args:
    ----
        data (dict): Specifies a dictionary with the data. The
            generated batches have the same structure that this input.
        batch_size (int): Specifies the batch size.
        shuffle (bool, optional): If ``True``, the examples are
            shuffled before to create the batches. Default: ``False``
        random_seed (int, optional): Specifies the random seed used to
            shuffle the data. Default: ``11918852809641073385``
    """

    def __init__(
        self,
        datapipe_or_data: IterDataPipe[dict[Hashable, Tensor]] | dict[Hashable, Tensor],
        batch_size: int,
        shuffle: bool = False,
        random_seed: int = 11918852809641073385,
    ) -> None:
        self._datapipe_or_data = datapipe_or_data
        self._batch_size = int(batch_size)
        self._shuffle = bool(shuffle)
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[dict]:
        datapipe_or_data = self._datapipe_or_data
        if not isinstance(datapipe_or_data, IterDataPipe):
            datapipe_or_data = SourceWrapperIterDataPipe([datapipe_or_data])
        for batch in datapipe_or_data:
            if self._shuffle:
                batch = shuffle_tensor_mapping(batch, generator=self._generator)
            keys = batch.keys()
            for tensors in zip(*[torch.split(value, self._batch_size) for value in batch.values()]):
                yield {key: tensor for key, tensor in zip(keys, tensors)}

    def __len__(self) -> int:
        if isinstance(self._datapipe_or_data, IterDataPipe):
            raise TypeError(f"{type(self).__qualname__} instance doesn't have valid length")
        return (
            get_first_value(self._datapipe_or_data).shape[0] + self._batch_size - 1
        ) // self._batch_size

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        desc = (
            str(self._datapipe_or_data)
            if isinstance(self._datapipe_or_data, IterDataPipe)
            else summary(self._datapipe_or_data)
        )
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  batch_size={self._batch_size},\n"
            f"  shuffle={self._shuffle},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe_or_data={str_indent(desc)},\n)"
        )

    def __getstate__(self) -> dict:
        state = super().__getstate__().copy()
        # torch.Generator cannot be serialized but its state can.
        state["_generator"] = state["_generator"].get_state()
        return state

    def __setstate__(self, state: dict) -> None:
        # Recreate the torch.Generator because only its state was serialized
        generator = torch.Generator()
        state["_generator"] = generator.set_state(state["_generator"])
        self.__dict__.update(state)

    @property
    def random_seed(self) -> int:
        r"""``int``: The random seed used to initialize the pseudo
        random generator."""
        return self._generator.initial_seed()


class TupleBatcherIterDataPipe(IterDataPipe[tuple[Tensor, ...]]):
    r"""Implements a source DataPipe to generate batch of examples from a
    tuple of ``torch.Tensor``s.

    Args:
    ----
        datapipe_or_data (``IterDataPipe`` or sequence of
            ``torch.Tensor`` of shape ``(num_examples, *)`` where ``*``
            means any number of dimensions): Specifies source DataPipe
            or a sequence of tensors.
        batch_size (int): Specifies the batch size.
        shuffle (bool, optional): If ``True``, the examples are
            shuffled before to create the batches. The
            shuffling is done per sequence of tensors.
            Default: ``False``
        random_seed (int, optional): Specifies the random seed used to
            shuffle the data. Default: ``13382866045483866228``
    """

    def __init__(
        self,
        datapipe_or_data: IterDataPipe[Sequence[Tensor]] | Sequence[Tensor],
        batch_size: int,
        shuffle: bool = False,
        random_seed: int = 13382866045483866228,
    ) -> None:
        self._datapipe_or_data = datapipe_or_data
        self._batch_size = int(batch_size)
        self._shuffle = bool(shuffle)
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[tuple[Tensor, ...]]:
        datapipe_or_tensors = self._datapipe_or_data
        if not isinstance(datapipe_or_tensors, IterDataPipe):
            datapipe_or_tensors = SourceWrapperIterDataPipe([datapipe_or_tensors])
        for batch in datapipe_or_tensors:
            if self._shuffle:
                batch = shuffle_tensors(batch, generator=self._generator)
            yield from zip(*[torch.split(tensor, self._batch_size) for tensor in batch])

    def __len__(self) -> int:
        if isinstance(self._datapipe_or_data, IterDataPipe):
            raise TypeError(f"{type(self).__qualname__} instance doesn't have valid length")
        return (self._datapipe_or_data[0].shape[0] + self._batch_size - 1) // self._batch_size

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        desc = (
            str(self._datapipe_or_data)
            if isinstance(self._datapipe_or_data, IterDataPipe)
            else summary(self._datapipe_or_data)
        )
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  batch_size={self._batch_size},\n"
            f"  shuffle={self._shuffle},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe_or_data={str_indent(desc)},\n)"
        )

    def __getstate__(self) -> dict:
        state = super().__getstate__().copy()
        # torch.Generator cannot be serialized but its state can.
        state["_generator"] = state["_generator"].get_state()
        return state

    def __setstate__(self, state: dict) -> None:
        # Recreate the torch.Generator because only its state was serialized
        generator = torch.Generator()
        state["_generator"] = generator.set_state(state["_generator"])
        self.__dict__.update(state)

    @property
    def random_seed(self) -> int:
        r"""``int``: The random seed used to initialize the pseudo
        random generator."""
        return self._generator.initial_seed()



r"""This module implements an asset manager."""

from __future__ import annotations

__all__ = ["AssetExistsError", "AssetManager", "AssetNotFoundError", "get_asset_summary"]

import copy
import logging
from typing import Any

import numpy as np
import torch
from coola import objects_are_equal

from gravitorch.utils.format import str_indent, str_torch_mapping

logger = logging.getLogger(__name__)


class AssetExistsError(Exception):
    r"""Raised when trying to add an asset that already exists."""


class AssetNotFoundError(Exception):
    r"""Raised when trying to access an asset that does not exist."""


class AssetManager:
    r"""Implements an asset manager."""

    def __init__(self, assets: dict[str, Any] | None = None) -> None:
        self._assets = assets or {}

    def __repr__(self) -> str:
        summaries = {name: self._get_asset_summary(asset) for name, asset in self._assets.items()}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_torch_mapping(summaries))}\n"
            ")"
        )

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(num_assets={len(self._assets):,})"

    def add_asset(self, name: str, asset: Any, replace_ok: bool = False) -> None:
        r"""Adds an asset to the asset manager.

        Note that the name should be unique. If the name exists, the
        old asset will be overwritten by the new asset.

        Args:
        ----
            name (str): Specifies the name of the asset to add.
            asset: Specifies the asset to add.
            replace_ok (bool, optional): If ``False``,
                ``AssetExistsError`` is raised if an asset with the
                same name exists. Default: ``False``

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
        """
        if name in self._assets and not replace_ok:
            raise AssetExistsError(
                f"`{name}` is already used to register an asset. "
                "Use `replace_ok=True` to replace an asset"
            )
        self._assets[name] = asset

    def clone(self) -> AssetManager:
        r"""Creates a deep copy of the current asset manager.

        Returns
        -------
            ``AssetManager``: A deep copy of the current asset manager.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager({"name": 5})
            >>> clone = manager.clone()
            >>> manager.add_asset("name", 7)
            >>> manager
            AssetManager(
              (name) <class 'int'>  7
            )
            >>> clone
            AssetManager(
              (name) <class 'int'>  5
            )
        """
        return AssetManager(copy.deepcopy(self._assets))

    def equal(self, other: Any) -> bool:
        r"""Indicates if two objects are equal.

        Args:
        ----
            other: Specifies the object to compare with.

        Returns:
        -------
            bool: ``True`` if the two objects are equal, otherwise
                ``False``.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager2 = AssetManager()
            >>> manager.equal(manager2)
            False
            >>> manager2.add_asset("mean", 5)
            >>> manager.equal(manager2)
            True
        """
        if not isinstance(other, AssetManager):
            return False
        return objects_are_equal(self._assets, other._assets)

    def get_asset(self, name: str) -> Any:
        r"""Gets an asset.

        Args:
        ----
            name (str): Specifies the asset to get.

        Returns:
        -------
            The asset

        Raises:
        ------
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.get_asset("mean")
            5
        """
        if name not in self._assets:
            raise AssetNotFoundError(f"The asset '{name}' does not exist")
        return self._assets[name]

    def get_asset_names(self) -> tuple[str, ...]:
        r"""Gets all the asset names.

        Returns
        -------
            tuple: The asset names.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.get_asset_names()
            ('mean',)
        """
        return tuple(self._assets.keys())

    def has_asset(self, name: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
        ----
            name (str): Specifies the name of the asset.

        Returns:
        -------
            bool: ``True`` if the asset exists, otherwise ``False``

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.has_asset("mean")
            False
            >>> manager.add_asset("mean", 5)
            >>> manager.has_asset("mean")
            True
        """
        return name in self._assets

    def remove_asset(self, name: str) -> None:
        r"""Removes an asset.

        Args:
        ----
            name (str): Specifies the name of the asset to remove.

        Raises:
        ------
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.utils.asset import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset("mean", 5)
            >>> manager.remove_asset("mean")
            5
            >>> manager.has_asset("mean")
            False
        """
        if name not in self._assets:
            raise AssetNotFoundError(
                f"The asset '{name}' does not exist so it is not possible to remove it"
            )
        del self._assets[name]

    def _get_asset_summary(self, asset: Any) -> str:
        return get_asset_summary(asset)


def get_asset_summary(asset: Any) -> str:
    r"""Gets a one line summary of an asset.

    Args:
    ----
        asset: Specifies the asset.

    Returns:
    -------
        str: A one line summary of an asset.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.utils.asset import get_asset_summary
        >>> get_asset_summary(torch.ones(2, 3))
        <class 'torch.Tensor'>  shape=torch.Size([2, 3])  dtype=torch.float32  device=cpu
    """
    if torch.is_tensor(asset):
        return f"{type(asset)}  shape={asset.shape}  dtype={asset.dtype}  device={asset.device}"
    if isinstance(asset, np.ndarray):
        return f"{type(asset)}  shape={asset.shape}  dtype={asset.dtype}"
    if isinstance(asset, (bool, int, float)):
        return f"{type(asset)}  {asset}"
    return f"{type(asset)}"



r"""This module implements some concatenation based fusion layers."""

from __future__ import annotations

__all__ = [
    "AverageFusion",
    "ConcatFusion",
    "FusionFFN",
    "MultiplicationFusion",
    "SumFusion",
]


import torch
from torch import Tensor
from torch.nn import Module

from gravitorch.nn.utils import get_module_output_size, setup_module


class ConcatFusion(Module):
    r"""Implements a module to concatenate inputs.

    Args:
    ----
        dim (int, optional): Specifies the fusion dimension. ``-1``
            means the last dimension. Default: ``-1``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.nn import ConcatFusion
        >>> module = ConcatFusion()
        >>> module
        ConcatFusion(dim=-1)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        tensor([[ 2.,  3.,  4., 12., 13., 14.],
                [ 5.,  6.,  7., 15., 16., 17.]], grad_fn=<CatBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, dim: int = -1) -> None:
        super().__init__()
        self._dim = dim

    def extra_repr(self) -> str:
        return f"dim={self._dim}"

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Concatenates the list or tuple of inputs and then applied a
        feed-forward network (FFN) on the fused representation.

        Args:
        ----
            *inputs (list or tuple of ``torch.Tensor``): Specifies the
                tensors to concatenate.

        Returns:
        -------
            ``torch.Tensor``: The fused representation.
        """
        if not inputs:
            raise RuntimeError(f"{self.__class__.__qualname__} needs at least one tensor as input")
        return torch.cat(inputs, dim=self._dim)


class FusionFFN(Module):
    r"""Implements a module that fuses representations and then applies a
    feed-forward network (FFN) on the fused representation.

    Args:
    ----
        fusion (``torch.nn.Module``, optional): Specifies the fusion
            module or its configuration.
        ffn (``torch.nn.Module``, optional): Specifies the FFN or its
            configuration.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.nn import ConcatFusion, FusionFFN
        >>> module = FusionFFN(ConcatFusion(), torch.nn.Linear(6, 4))
        >>> module
        FusionFFN(
          (fusion): ConcatFusion(dim=-1)
          (ffn): Linear(in_features=6, out_features=4, bias=True)
        )
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 2.1302,  7.8643,  8.3265,  4.7277],
                [ 1.6980, 11.7446,  9.8449,  5.8402]], grad_fn=<AddmmBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, fusion: Module | dict, ffn: Module | dict) -> None:
        super().__init__()
        self.fusion = setup_module(fusion)
        self.ffn = setup_module(ffn)

    @property
    def output_size(self) -> int:
        r"""``int``: The output size of the module."""
        return get_module_output_size(self.ffn)

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Fuses the inputs and then applied a feed-forward network
        (FFN) on the fused representation.

        Args:
        ----
            *inputs (sequence of ``torch.Tensor``): Specifies the
                sequence of tensors to fuse. The shape of the tensors
                may depend on the feed-forward network (FFN).

        Returns:
        -------
            ``torch.Tensor``: The fused representation.
        """
        return self.ffn(self.fusion(*inputs))


class MultiplicationFusion(Module):
    r"""Defines a fusion layer that multiplies the inputs.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.nn import MultiplicationFusion
        >>> module = MultiplicationFusion()
        >>> module
        MultiplicationFusion()
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 24.,  39.,  56.],
                [ 75.,  96., 119.]], grad_fn=<MulBackward0>)
        >>> out.mean().backward()
    """

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Multiplies the list or tuple of inputs.

        Args:
        ----
            *inputs (list or tuple of tensors): Specifies the list or
                tuple of tensors to multiply. The shape of the tensors
                should be the same. By default, this layer expects
                that each input is a ``torch.Tensor`` of shape
                ``(batch size, feature size)``. But it can also work
                if the inputs have a shape
                ``(sequence length, batch size, feature size)`` or
                similar shapes.

        Returns:
        -------
            ``torch.Tensor`` with the same shape that the input
                tensor: The fused tensor.
        """
        if not inputs:
            raise RuntimeError(f"{self.__class__.__qualname__} needs at least one tensor as input")
        output = inputs[0]
        for xi in inputs[1:]:
            output = output.mul(xi)
        return output


class SumFusion(Module):
    r"""Defines a layer to sum the inputs.

    Args:
    ----
        normalized (bool, optional): Specifies the output is
            normalized by the number of inputs.
            Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.nn import SumFusion
        >>> module = SumFusion()
        >>> module
        SumFusion(normalized=False)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[14., 16., 18.],
                [20., 22., 24.]], grad_fn=<AddBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, normalized: bool = False) -> None:
        super().__init__()
        self._normalized = bool(normalized)

    def extra_repr(self) -> str:
        return f"normalized={self._normalized}"

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Sums the list or tuple of inputs.

        Args:
        ----
            *inputs (list or tuple of tensors): Specifies the list or
                tuple of tensors to sum. The shape of the tensors
                should be the same. By default, this layer expects
                that each input is a ``torch.Tensor`` of shape
                ``(batch size, feature size)``. But it can also work
                if the inputs have a shape
                ``(sequence length, batch size, feature size)`` or
                similar shapes.

        Returns:
        -------
            ``torch.Tensor`` with the same shape that the input
                tensor: The fused tensor.
        """
        if not inputs:
            raise RuntimeError(f"{self.__class__.__qualname__} needs at least one tensor as input")

        output = inputs[0]
        for x in inputs[1:]:
            output = output + x

        if self._normalized:
            output = output.div(len(inputs))
        return output


class AverageFusion(SumFusion):
    r"""Implements a layer to average the inputs.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.nn import AverageFusion
        >>> module = AverageFusion()
        >>> module
        AverageFusion(normalized=True)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 7.,  8.,  9.],
                [10., 11., 12.]], grad_fn=<DivBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self) -> None:
        super().__init__(normalized=True)



r"""This module implements some utility functions for the training
loops."""

__all__ = ["is_training_loop_config", "setup_training_loop"]

import logging
from typing import Union

from objectory.utils import is_object_config

from gravitorch.loops.training.base import BaseTrainingLoop
from gravitorch.loops.training.vanilla import VanillaTrainingLoop
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_training_loop_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseTrainingLoop``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseTrainingLoop`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.training import is_training_loop_config
        >>> is_training_loop_config({"_target_": "gravitorch.loops.training.VanillaTrainingLoop"})
        True
    """
    return is_object_config(config, BaseTrainingLoop)


def setup_training_loop(training_loop: Union[BaseTrainingLoop, dict, None]) -> BaseTrainingLoop:
    r"""Sets up the training loop.

    The training loop is instantiated from its configuration by using
    the ``BaseTrainingLoop`` factory function.

    Args:
    ----
        training_loop (``BaseTrainingLoop`` or dict or None):
            Specifies the training loop or its configuration.
            If ``None``, the ``VanillaTrainingLoop`` is instantiated.

    Returns:
    -------
        ``BaseTrainingLoop``: The training loop.
    """
    if training_loop is None:
        training_loop = VanillaTrainingLoop()
    if isinstance(training_loop, dict):
        logger.info(
            "Initializing a training loop from its configuration... "
            f"{str_target_object(training_loop)}"
        )
        training_loop = BaseTrainingLoop.factory(**training_loop)
    if not isinstance(training_loop, BaseTrainingLoop):
        logger.warning(
            f"training_loop is not a `BaseTrainingLoop` (received: {type(training_loop)})"
        )
    return training_loop



__all__ = [
    "BaseLoopObserver",
    "NoOpLoopObserver",
    "PyTorchBatchSaver",
    "SequentialLoopObserver",
    "is_loop_observer_config",
    "setup_loop_observer",
]

from gravitorch.loops.observers.base import BaseLoopObserver
from gravitorch.loops.observers.batch_saving import PyTorchBatchSaver
from gravitorch.loops.observers.factory import (
    is_loop_observer_config,
    setup_loop_observer,
)
from gravitorch.loops.observers.noop import NoOpLoopObserver
from gravitorch.loops.observers.sequential import SequentialLoopObserver



from __future__ import annotations

__all__ = ["SequentialRunner"]

import logging
from collections.abc import Sequence
from typing import Any

from gravitorch.runners.base import BaseRunner
from gravitorch.runners.utils import setup_runner
from gravitorch.utils.format import str_indent, str_torch_sequence

logger = logging.getLogger(__name__)


class SequentialRunner(BaseRunner):
    r"""Implements a runner that executes multiple runners sequentially.

    Args:
    ----
        runners (sequence): Specifies the sequence of runners or their
            configuration.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.runners import TrainingRunner, SequentialRunner, EvaluationRunner
        >>> engine = ...
        >>> runner = SequentialRunner([TrainingRunner(engine), EvaluationRunner(engine)])
        >>> runner.run()
    """

    def __init__(self, runners: Sequence[BaseRunner | dict]) -> None:
        self._runners = tuple(setup_runner(runner) for runner in runners)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_torch_sequence(self._runners))}\n)"
        )

    def run(self) -> Any:
        for runner in self._runners:
            runner.run()
